"""Golden tests cases for testing liquid's `tablerow` tag."""

from liquid.golden.case import Case

cases = [
    Case(
        description="one row",
        template=r"{% tablerow tag in collection.tags %}{{ tag }}{% endtablerow %}",
        expect=(
            '<tr class="row1">\n'
            '<td class="col1">tag1</td>'
            '<td class="col2">tag2</td>'
            '<td class="col3">tag3</td>'
            '<td class="col4">tag4</td>'
            "</tr>\n"
        ),
        globals={
            "collection": {
                "tags": [
                    "tag1",
                    "tag2",
                    "tag3",
                    "tag4",
                ]
            }
        },
    ),
    Case(
        description="one row with limit",
        template=(
            r"{% tablerow tag in collection.tags limit: 2 %}"
            r"{{ tag }}"
            r"{% endtablerow %}"
        ),
        expect=(
            '<tr class="row1">\n'
            '<td class="col1">tag1</td>'
            '<td class="col2">tag2</td>'
            "</tr>\n"
        ),
        globals={
            "collection": {
                "tags": [
                    "tag1",
                    "tag2",
                    "tag3",
                    "tag4",
                ]
            }
        },
    ),
    Case(
        description="one row with offset",
        template=(
            r"{% tablerow tag in collection.tags offset: 2 %}"
            r"{{ tag }}"
            r"{% endtablerow %}"
        ),
        expect=(
            '<tr class="row1">\n'
            '<td class="col1">tag3</td>'
            '<td class="col2">tag4</td>'
            "</tr>\n"
        ),
        globals={
            "collection": {
                "tags": [
                    "tag1",
                    "tag2",
                    "tag3",
                    "tag4",
                ]
            }
        },
    ),
    Case(
        description="two columns",
        template=(
            r"{% tablerow tag in collection.tags cols:2 %}"
            r"{{ tag }}"
            r"{% endtablerow %}"
        ),
        expect=(
            '<tr class="row1">\n'
            '<td class="col1">tag1</td>'
            '<td class="col2">tag2</td>'
            "</tr>\n"
            '<tr class="row2">'
            '<td class="col1">tag3</td>'
            '<td class="col2">tag4</td>'
            "</tr>\n"
        ),
        globals={
            "collection": {
                "tags": [
                    "tag1",
                    "tag2",
                    "tag3",
                    "tag4",
                ]
            }
        },
    ),
    Case(
        description="two column range",
        template=(
            r"{% tablerow i in (1..4) cols:2 %}"
            r"{{ i }} {{ tablerowloop.col_first }}"
            r"{% endtablerow %}"
        ),
        expect=(
            '<tr class="row1">\n'
            '<td class="col1">1 true</td>'
            '<td class="col2">2 false</td>'
            "</tr>\n"
            '<tr class="row2">'
            '<td class="col1">3 true</td>'
            '<td class="col2">4 false</td>'
            "</tr>\n"
        ),
    ),
    Case(
        description="two column odd range",
        template=(
            r"{% tablerow i in (1..5) cols:2 %}"
            r"{{ i }} {{ tablerowloop.col_first }}"
            r"{% endtablerow %}"
        ),
        expect=(
            '<tr class="row1">\n'
            '<td class="col1">1 true</td>'
            '<td class="col2">2 false</td>'
            "</tr>\n"
            '<tr class="row2">'
            '<td class="col1">3 true</td>'
            '<td class="col2">4 false</td>'
            "</tr>\n"
            '<tr class="row3">'
            '<td class="col1">5 true</td>'
            "</tr>\n"
        ),
    ),
    Case(
        description="two column odd range row numbers",
        template=(
            r"{% tablerow i in (1..5) cols:2 %}"
            r"{{ i }} {{ tablerowloop.row }}"
            r"{% endtablerow %}"
        ),
        expect=(
            '<tr class="row1">\n'
            '<td class="col1">1 1</td>'
            '<td class="col2">2 1</td>'
            "</tr>\n"
            '<tr class="row2">'
            '<td class="col1">3 2</td>'
            '<td class="col2">4 2</td>'
            "</tr>\n"
            '<tr class="row3">'
            '<td class="col1">5 3</td>'
            "</tr>\n"
        ),
    ),
    Case(
        description="no cols param",
        template=(
            "\n".join(
                [
                    "{% tablerow i in (1..2) %}",
                    "col: {{ tablerowloop.col }}",
                    "col0: {{ tablerowloop.col0 }}",
                    "col_first: {{ tablerowloop.col_first }}",
                    "col_last: {{ tablerowloop.col_last }}",
                    "first: {{ tablerowloop.first }}",
                    "index: {{ tablerowloop.index }}",
                    "index0: {{ tablerowloop.index0 }}",
                    "last: {{ tablerowloop.last }}",
                    "length: {{ tablerowloop.length }}",
                    "rindex: {{ tablerowloop.rindex }}",
                    "rindex0: {{ tablerowloop.rindex0 }}",
                    "row: {{ tablerowloop.row }}",
                    "{% endtablerow %}",
                ]
            )
        ),
        expect=(
            "\n".join(
                [
                    '<tr class="row1">',
                    '<td class="col1">',
                    "col: 1",
                    "col0: 0",
                    "col_first: true",
                    "col_last: false",
                    "first: true",
                    "index: 1",
                    "index0: 0",
                    "last: false",
                    "length: 2",
                    "rindex: 2",
                    "rindex0: 1",
                    "row: 1",
                    '</td><td class="col2">',
                    "col: 2",
                    "col0: 1",
                    "col_first: false",
                    "col_last: true",
                    "first: false",
                    "index: 2",
                    "index0: 1",
                    "last: true",
                    "length: 2",
                    "rindex: 1",
                    "rindex0: 0",
                    "row: 1",
                    "</td></tr>\n",
                ]
            )
        ),
    ),
    Case(
        description="cols is a string",
        template=(
            r"{% tablerow i in (1..4) cols:'2' %}"
            r"{{ i }} {{ tablerowloop.col_first }}"
            r"{% endtablerow %}"
        ),
        expect=(
            '<tr class="row1">\n'
            '<td class="col1">1 true</td>'
            '<td class="col2">2 false</td>'
            "</tr>\n"
            '<tr class="row2">'
            '<td class="col1">3 true</td>'
            '<td class="col2">4 false</td>'
            "</tr>\n"
        ),
    ),
    Case(
        description="cols is a float",
        template=(
            r"{% tablerow i in (1..4) cols:2.6 %}"
            r"{{ i }} {{ tablerowloop.col_first }}"
            r"{% endtablerow %}"
        ),
        expect=(
            '<tr class="row1">\n'
            '<td class="col1">1 true</td>'
            '<td class="col2">2 false</td>'
            "</tr>\n"
            '<tr class="row2">'
            '<td class="col1">3 true</td>'
            '<td class="col2">4 false</td>'
            "</tr>\n"
        ),
    ),
    Case(
        description="limit is a string",
        template=(
            r"{% tablerow i in (1..4) limit:'2' %}"
            r"{{ i }} {{ tablerowloop.col_first }}"
            r"{% endtablerow %}"
        ),
        expect=(
            '<tr class="row1">\n'
            '<td class="col1">1 true</td>'
            '<td class="col2">2 false</td>'
            "</tr>\n"
        ),
    ),
    Case(
        description="offset is a string",
        template=(
            r"{% tablerow i in (1..4) offset:'2' %}"
            r"{{ i }} {{ tablerowloop.col_first }}"
            r"{% endtablerow %}"
        ),
        expect=(
            '<tr class="row1">\n'
            '<td class="col1">3 true</td>'
            '<td class="col2">4 false</td>'
            "</tr>\n"
        ),
    ),
    # Case(
    #     description="cols is non number string",
    #     template=(
    #         r"{% tablerow i in (1..4) cols:'foo' %}"
    #         r"{{ i }} {{ tablerowloop.col_first }}"
    #         r"{% endtablerow %}"
    #     ),
    #     expect=(
    #         '<tr class="row1">\n'
    #         '<td class="col1">1 true</td>'
    #         '<td class="col2">2 false</td>'
    #         '<td class="col3">3 false</td>'
    #         '<td class="col4">4 false</td>'
    #         "</tr>\n"
    #     ),
    # ),
    # Case(
    #     description="limit is non number string",
    #     template=(
    #         r"{% tablerow i in (1..4) limit:'foo' %}"
    #         r"{{ i }} {{ tablerowloop.col_first }}"
    #         r"{% endtablerow %}"
    #     ),
    #     expect='<tr class="row1">\n</tr>\n',
    # ),
    # Case(
    #     description="offset is non number string",
    #     template=(
    #         r"{% tablerow i in (1..4) offset:'foo' %}"
    #         r"{{ i }} {{ tablerowloop.col_first }}"
    #         r"{% endtablerow %}"
    #     ),
    #     expect=(
    #         '<tr class="row1">\n'
    #         '<td class="col1">1 true</td>'
    #         '<td class="col2">2 false</td>'
    #         '<td class="col3">3 false</td>'
    #         '<td class="col4">4 false</td>'
    #         "</tr>\n"
    #     ),
    # ),
    # Case(
    #     description="offset is an array",
    #     template=(
    #         r"{% tablerow i in (1..4) offset:foo %}"
    #         r"{{ i }} {{ tablerowloop.col_first }}"
    #         r"{% endtablerow %}"
    #     ),
    #     expect=(
    #         '<tr class="row1">\n'
    #         '<td class="col1">1 true</td>'
    #         '<td class="col2">2 false</td>'
    #         '<td class="col3">3 false</td>'
    #         '<td class="col4">4 false</td>'
    #         "</tr>\n"
    #     ),
    #     globals={"foo": [1, 2, 3]},
    #     error=True,
    # ),
]



"""Filter functions that operate on arrays."""
from __future__ import annotations

from functools import partial
from itertools import chain
from itertools import islice
from operator import getitem

from typing import Any
from typing import Iterable
from typing import List
from typing import Sequence
from typing import Tuple
from typing import Union
from typing import TYPE_CHECKING

try:
    from markupsafe import Markup
except ImportError:
    from liquid.exceptions import Markup  # type: ignore

from liquid.filter import with_environment
from liquid.filter import array_filter
from liquid.filter import sequence_filter
from liquid.filter import liquid_filter

from liquid.expression import NIL

from liquid.exceptions import FilterError
from liquid.exceptions import FilterArgumentError

from liquid import is_undefined

if TYPE_CHECKING:
    from liquid import Environment

ArrayT = Union[List[Any], Tuple[Any, ...]]

# Send objects with missing keys to the end when sorting a list.
MAX_CH = chr(0x10FFFF)

# Unique object for use with the uniq filter.
MISSING = object()


def _str_if_not(val: object) -> str:
    if not isinstance(val, str):
        return str(val)
    return val


def _getitem(sequence: Any, key: object, default: object = None) -> Any:
    """Helper for the map filter.

    Same as sequence[key], but returns a default value if key does not exist
    in sequence.
    """
    try:
        return getitem(sequence, key)
    except (KeyError, IndexError):
        return default
    except TypeError:
        if not hasattr(sequence, "__getitem__"):
            raise
        return default


def _lower(obj: Any) -> str:
    """Helper for the sort filter"""
    try:
        return str(obj).lower()
    except AttributeError:
        return ""


@with_environment
@sequence_filter
def join(
    sequence: Iterable[object], separator: object = " ", *, environment: Environment
) -> str:
    """Concatenate an array of strings."""
    if not isinstance(separator, str):
        separator = str(separator)

    if environment.autoescape and separator == " ":
        separator = Markup(" ")

    return separator.join(_str_if_not(item) for item in sequence)


@liquid_filter
def first(obj: Any) -> object:
    """Return the first item of an array"""
    if isinstance(obj, str):
        return None

    if isinstance(obj, dict):
        obj = list(islice(obj.items(), 1))

    try:
        return getitem(obj, 0)
    except (TypeError, KeyError, IndexError):
        return None


@liquid_filter
def last(obj: Sequence[Any]) -> object:
    """Return the last item of an array"""
    if isinstance(obj, str):
        return None

    try:
        return getitem(obj, -1)
    except (TypeError, KeyError, IndexError):
        return None


@sequence_filter
def concat(sequence: ArrayT, second_array: ArrayT) -> ArrayT:
    """Return two arrays joined together."""
    if not isinstance(second_array, (list, tuple)):
        raise FilterArgumentError(
            f"concat expected an array, found {type(second_array).__name__}"
        )

    if is_undefined(sequence):
        return second_array

    return list(chain(sequence, second_array))


@liquid_filter
def map_(sequence: ArrayT, key: object) -> List[object]:
    """Creates an array of values by extracting the values of a named property
    from another object."""
    try:
        return [_getitem(itm, str(key), default=NIL) for itm in sequence]
    except TypeError as err:
        raise FilterError("can't map sequence") from err


@array_filter
def reverse(array: ArrayT) -> List[object]:
    """Reverses the order of the items in an array."""
    return list(reversed(array))


@array_filter
def sort(sequence: ArrayT, key: object = None) -> List[object]:
    """Sorts items in an array in case-sensitive order.

    When a key string is provided, objects without the key property should
    be at the end of the output list/array.
    """
    if key:
        key_func = partial(_getitem, key=str(key), default=MAX_CH)
        return list(sorted(sequence, key=key_func))

    try:
        return list(sorted(sequence))
    except TypeError as err:
        raise FilterError("can't sort sequence") from err


@array_filter
def sort_natural(sequence: ArrayT, key: object = None) -> List[object]:
    """Sorts items in an array in case-insensitive order."""
    if key:
        item_getter = partial(_getitem, key=str(key), default=MAX_CH)
        return list(sorted(sequence, key=lambda obj: _lower(item_getter(obj))))

    return list(sorted(sequence, key=_lower))


@sequence_filter
def where(sequence: ArrayT, attr: object, value: object = None) -> List[object]:
    """Creates an array including only the objects with a given property value,
    or any truthy value by default."""
    if value is not None and not is_undefined(value):
        return [itm for itm in sequence if _getitem(itm, attr) == value]

    return [itm for itm in sequence if _getitem(itm, attr) not in (False, None)]


@sequence_filter
def uniq(sequence: ArrayT, key: object = None) -> List[object]:
    """Removes any duplicate elements in an array."""
    # Note that we're not using a dict or set for deduplication because we need
    # to handle sequences containing unhashable objects, like dictionaries.

    # This is probably quite slow.
    if key is not None:
        keys = []
        result = []
        for obj in sequence:
            try:
                item = obj[key]
            except KeyError:
                item = MISSING
            except TypeError as err:
                raise FilterArgumentError(
                    f"can't read property '{key}' of {obj}"
                ) from err

            if item not in keys:
                keys.append(item)
                result.append(obj)

        return result

    return [obj for i, obj in enumerate(sequence) if sequence.index(obj) == i]


@sequence_filter
def compact(sequence: ArrayT, key: object = None) -> List[object]:
    """Removes any nil values from an array."""
    if key is not None:
        try:
            return [itm for itm in sequence if itm[key] is not None]
        except TypeError as err:
            raise FilterArgumentError(f"can't read property '{key}'") from err
    return [itm for itm in sequence if itm is not None]



"""Golden test case definition."""
from dataclasses import dataclass
from dataclasses import field

from typing import Any
from typing import Dict


# pylint: disable=too-many-instance-attributes
@dataclass
class Case:
    """Test case dataclass to help with table driven tests."""

    description: str
    template: str
    expect: str
    globals: Dict[str, Any] = field(default_factory=dict)
    partials: Dict[str, Any] = field(default_factory=dict)
    standard: bool = True
    error: bool = False
    strict: bool = False


TEMPLATE_DROP_ATTRS = (
    r"{{ template }} "
    r"{{ template.directory }} "
    r"{{ template.name }} "
    r"{{ template.suffix }}"
)



"""Golden tests cases for testing liquid's `#` tag."""

from liquid.golden.case import Case

cases = [
    Case(
        description="with whitespace control and padding",
        template=r"{%- # some comment -%}",
        expect="",
    ),
    Case(
        description="with whitespace control no padding",
        template=r"{%-# some comment -%}",
        expect="",
    ),
    Case(
        description="no whitespace control with padding",
        template=r"{% # some comment %}",
        expect="",
    ),
    Case(
        description="no whitespace control no padding",
        template=r"{%# some comment %}",
        expect="",
    ),
    Case(
        description="no padding after the hash",
        template=r"{%#some comment %}",
        expect="",
    ),
    Case(
        description="empty",
        template=r"{%#%}",
        expect="",
    ),
    Case(
        description="liquid tag",
        template="\n".join(
            [
                r"{% liquid ",
                r"  # first comment line",
                r"  # second comment line",
                r"",
                r"  # another comment line",
                r"  echo 'Hello '",
                r"",
                r"  # more comments",
                r"  echo 'goodbye'",
                r"-%}",
            ]
        ),
        expect="Hello goodbye",
    ),
    Case(
        description="multiple lines",
        template="\n".join(
            [
                "{%-",
                "  # spread inline comments",
                "  # over multiple lines",
                "-%}",
            ]
        ),
        expect="",
    ),
    Case(
        description="lots of hashes in a liquid tag",
        template="\n".join(
            [
                r"{% liquid",
                r"  ##########################",
                r"  # spread inline comments #",
                r"  ##########################",
                r"-%}",
            ]
        ),
        expect="",
    ),
    Case(
        description="enforce leading hash",
        template="\n".join(
            [
                "{%-",
                "  # spread inline comments",
                "  over multiple lines",
                "-%}",
            ]
        ),
        expect="",
        error=True,
    ),
    Case(
        description="can't comment tags",
        template="{%- # {% echo 'hello world' %} -%}",
        expect=" -%}",
    ),
]



"""Test loop-style expression tokenization.

Loop-style expressions are those found in the `for` tag and
the `tablerow` tag.
"""

import unittest
from typing import Any
from typing import NamedTuple

from liquid.expressions.loop import tokenize

from liquid.token import TOKEN_IN
from liquid.token import TOKEN_COLON
from liquid.token import TOKEN_FLOAT
from liquid.token import TOKEN_DOT
from liquid.token import TOKEN_REVERSED
from liquid.token import TOKEN_IDENTIFIER
from liquid.token import TOKEN_LIMIT
from liquid.token import TOKEN_INTEGER
from liquid.token import TOKEN_LPAREN
from liquid.token import TOKEN_RANGE
from liquid.token import TOKEN_RPAREN
from liquid.token import TOKEN_OFFSET
from liquid.token import TOKEN_CONTINUE
from liquid.token import TOKEN_COMMA
from liquid.token import Token


class Case(NamedTuple):
    """Table-driven test helper."""

    description: str
    source: str
    expect: Any


class LexIncludeExpressionTestCase(unittest.TestCase):
    """Test cases for `include` expressions."""

    def test_lex_include_expression(self) -> None:
        """Test that we can tokenize `include` expressions."""
        test_cases = [
            Case(
                "loop over identifier",
                "product in collection.products",
                [
                    Token(1, TOKEN_IDENTIFIER, "product"),
                    Token(1, TOKEN_IN, "in"),
                    Token(1, TOKEN_IDENTIFIER, "collection"),
                    Token(1, TOKEN_DOT, "."),
                    Token(1, TOKEN_IDENTIFIER, "products"),
                ],
            ),
            Case(
                "loop over identifier with limit and offset",
                "product in collection.products limit:4 offset:2",
                [
                    Token(1, TOKEN_IDENTIFIER, "product"),
                    Token(1, TOKEN_IN, "in"),
                    Token(1, TOKEN_IDENTIFIER, "collection"),
                    Token(1, TOKEN_DOT, "."),
                    Token(1, TOKEN_IDENTIFIER, "products"),
                    Token(1, TOKEN_LIMIT, "limit"),
                    Token(1, TOKEN_COLON, ":"),
                    Token(1, TOKEN_INTEGER, "4"),
                    Token(1, TOKEN_OFFSET, "offset"),
                    Token(1, TOKEN_COLON, ":"),
                    Token(1, TOKEN_INTEGER, "2"),
                ],
            ),
            Case(
                "loop over reversed range",
                "num in (1..10) reversed",
                [
                    Token(1, TOKEN_IDENTIFIER, "num"),
                    Token(1, TOKEN_IN, "in"),
                    Token(1, TOKEN_LPAREN, "("),
                    Token(1, TOKEN_INTEGER, "1"),
                    Token(1, TOKEN_RANGE, ".."),
                    Token(1, TOKEN_INTEGER, "10"),
                    Token(1, TOKEN_RPAREN, ")"),
                    Token(1, TOKEN_REVERSED, "reversed"),
                ],
            ),
            Case(
                "loop over range with identifier",
                "i in (1..num)",
                [
                    Token(1, TOKEN_IDENTIFIER, "i"),
                    Token(1, TOKEN_IN, "in"),
                    Token(1, TOKEN_LPAREN, "("),
                    Token(1, TOKEN_INTEGER, "1"),
                    Token(1, TOKEN_RANGE, ".."),
                    Token(1, TOKEN_IDENTIFIER, "num"),
                    Token(1, TOKEN_RPAREN, ")"),
                ],
            ),
            Case(
                "loop over range with float start",
                "i in (2.4..5)",
                [
                    Token(1, TOKEN_IDENTIFIER, "i"),
                    Token(1, TOKEN_IN, "in"),
                    Token(1, TOKEN_LPAREN, "("),
                    Token(1, TOKEN_FLOAT, "2.4"),
                    Token(1, TOKEN_RANGE, ".."),
                    Token(1, TOKEN_INTEGER, "5"),
                    Token(1, TOKEN_RPAREN, ")"),
                ],
            ),
            Case(
                description="loop over named iterable with continue offset",
                source="item in array limit: 3 offset: continue",
                expect=[
                    Token(1, TOKEN_IDENTIFIER, "item"),
                    Token(1, TOKEN_IN, "in"),
                    Token(1, TOKEN_IDENTIFIER, "array"),
                    Token(1, TOKEN_LIMIT, "limit"),
                    Token(1, TOKEN_COLON, ":"),
                    Token(1, TOKEN_INTEGER, "3"),
                    Token(1, TOKEN_OFFSET, "offset"),
                    Token(1, TOKEN_COLON, ":"),
                    Token(1, TOKEN_CONTINUE, "continue"),
                ],
            ),
            Case(
                description="comma separated arguments",
                source="i in array, limit: 4, offset: 2",
                expect=[
                    Token(1, TOKEN_IDENTIFIER, "i"),
                    Token(1, TOKEN_IN, "in"),
                    Token(1, TOKEN_IDENTIFIER, "array"),
                    Token(1, TOKEN_COMMA, ","),
                    Token(1, TOKEN_LIMIT, "limit"),
                    Token(1, TOKEN_COLON, ":"),
                    Token(1, TOKEN_INTEGER, "4"),
                    Token(1, TOKEN_COMMA, ","),
                    Token(1, TOKEN_OFFSET, "offset"),
                    Token(1, TOKEN_COLON, ":"),
                    Token(1, TOKEN_INTEGER, "2"),
                ],
            ),
        ]

        for case in test_cases:
            with self.subTest(msg=case.description):
                tokens = list(tokenize(case.source))
                self.assertEqual(len(tokens), len(case.expect))

                for got, want in zip(tokens, case.expect):
                    self.assertEqual(got, want)



"""Expression caching test cases."""
# pylint: disable=missing-class-docstring

from unittest import TestCase

from liquid import Environment
from liquid import Template

from liquid.builtin.tags.if_tag import IfNode


class ExpressionCacheTestCase(TestCase):
    def test_env_with_no_expression_caching(self):
        """Test that there is no expression caching by default."""
        env = Environment(expression_cache_size=0)
        template = env.from_string(
            "{% if x == 1 %}{% endif %}{% if x == 1 %}{% endif %}"
        )
        nodes = template.tree.statements
        self.assertEqual(len(nodes), 2)

        node1, node2 = nodes
        assert isinstance(node1, IfNode)
        assert isinstance(node2, IfNode)

        self.assertNotEqual(id(node1.condition), id(node2.condition))

    def test_env_with_expression_caching(self):
        """Test that there is no expression caching by default."""
        env = Environment(expression_cache_size=1)
        template = env.from_string(
            "{% if x == 1 %}{% endif %}{% if x == 1 %}{% endif %}"
        )
        nodes = template.tree.statements
        self.assertEqual(len(nodes), 2)

        node1, node2 = nodes
        assert isinstance(node1, IfNode)
        assert isinstance(node2, IfNode)

        self.assertEqual(id(node1.condition), id(node2.condition))

    def test_template_with_no_expression_caching(self):
        """Test that there is no expression caching by default."""
        template = Template(
            "{% if x == 1 %}{% endif %}{% if x == 1 %}{% endif %}",
            expression_cache_size=0,
        )

        nodes = template.tree.statements
        self.assertEqual(len(nodes), 2)

        node1, node2 = nodes
        assert isinstance(node1, IfNode)
        assert isinstance(node2, IfNode)

        self.assertNotEqual(id(node1.condition), id(node2.condition))

    def test_template_with_expression_caching(self):
        """Test that there is no expression caching by default."""
        template = Template(
            "{% if x == 1 %}{% endif %}{% if x == 1 %}{% endif %}",
            expression_cache_size=1,
        )

        nodes = template.tree.statements
        self.assertEqual(len(nodes), 2)

        node1, node2 = nodes
        assert isinstance(node1, IfNode)
        assert isinstance(node2, IfNode)

        self.assertEqual(id(node1.condition), id(node2.condition))



"""Generic argument list parsers."""
from typing import Any
from typing import Callable
from typing import Dict
from typing import List
from typing import Optional
from typing import Tuple

from liquid.expression import Expression
from liquid.expression import NIL

from liquid.expressions.common import parse_string_literal
from liquid.expressions.common import parse_unchained_identifier
from liquid.expressions.arguments.lex import tokenize
from liquid.expressions.filtered.parse import parse_obj
from liquid.expressions.stream import TokenStream

from liquid.token import TOKEN_COLON
from liquid.token import TOKEN_COMMA
from liquid.token import TOKEN_EOF
from liquid.token import TOKEN_EQUALS


Argument = Tuple[Any, Expression]


def make_parse_arguments(
    func: Callable[[TokenStream, str], Argument],
    separator_token: str = TOKEN_COLON,
) -> Callable[[TokenStream], List[Argument]]:
    """Return an argument list parser using the given function to parse
    each argument."""

    def _parse_arguments(stream: TokenStream) -> List[Argument]:
        """Parse arguments from a stream of tokens until EOF."""
        args: List[Argument] = []
        # Leading commas are OK
        if stream.current[1] == TOKEN_COMMA:
            next(stream)

        while stream.current[1] != TOKEN_EOF:
            args.append(func(stream, separator_token))
            next(stream)
            # Catch missing comma
            if stream.current[1] != TOKEN_EOF:
                stream.expect(TOKEN_COMMA)
            # Eat comma. Trailing commas are OK
            if stream.current[1] == TOKEN_COMMA:
                next(stream)

        return args

    return _parse_arguments


def parse_argument(stream: TokenStream, separator_token: str) -> Argument:
    """Parse a single keyword argument from a stream of tokens."""
    key = str(parse_unchained_identifier(stream))
    next(stream)
    stream.expect(separator_token)
    next(stream)  # Eat separator
    return key, parse_obj(stream)


parse_colon_separated_arguments = make_parse_arguments(parse_argument, TOKEN_COLON)
parse_equals_separated_arguments = make_parse_arguments(parse_argument, TOKEN_EQUALS)


def parse_keyword_arguments(expr: str, linenum: int = 1) -> Dict[str, Expression]:
    """Parse keyword or named arguments from a Liquid expression string.

    Each key/value pair is assumed to be separated by a comma. Leading and
    trailing commas are OK.

    If the same key/name appears multiple times, the last occurrence in the
    argument "list" will take priority.

    Values can be string, integer, float, true, false or nil literals, an
    identifier or a range expression. An identifier value could be chained
    using a mixture of dot and bracket notation."""
    return dict(parse_colon_separated_arguments(TokenStream(tokenize(expr, linenum))))


def parse_macro_argument(stream: TokenStream, separator_token: str) -> Argument:
    """Parse either a positional or keyword argument.

    If the separator token appears after an identifier, the argument is a keyword
    argument. Otherwise it is positional, with an expression of NIL.
    """
    key = str(parse_unchained_identifier(stream))
    if stream.peek[1] == separator_token:
        next(stream)
        # A keyword argument
        next(stream)  # Eat separator
        val = parse_obj(stream)
    else:
        # A positional argument
        val = NIL
    return key, val


def parse_call_argument(stream: TokenStream, separator_token: str) -> Argument:
    """Parse an argument value that could be preceded by a name."""
    if stream.peek[1] == separator_token:
        # A keyword argument
        key: Optional[str] = str(parse_unchained_identifier(stream))
        next(stream)
        next(stream)  # Eat separator
    else:
        # A positional argument
        key = None

    return key, parse_obj(stream)


_parse_macro_arguments = make_parse_arguments(parse_macro_argument, TOKEN_COLON)
_parse_call_arguments = make_parse_arguments(parse_call_argument, TOKEN_COLON)


def parse_macro_arguments(expr: str, linenum: int = 1) -> Tuple[str, List[Argument]]:
    """Parse a sequence of argument names, possibly with default values."""
    stream = TokenStream(tokenize(expr, linenum))
    name = parse_string_literal(stream).value
    next(stream)
    return name, _parse_macro_arguments(stream)


def parse_call_arguments(expr: str, linenum: int = 1) -> Tuple[str, List[Argument]]:
    """Parse a sequence of positional and/or keyword arguments."""
    stream = TokenStream(tokenize(expr, linenum))
    name = parse_string_literal(stream).value
    next(stream)
    return name, _parse_call_arguments(stream)



"""Maths related filter definitions."""

import math
import decimal

from typing import Optional
from typing import Union

from liquid.context import is_undefined
from liquid.exceptions import FilterArgumentError

from liquid.filter import math_filter
from liquid.filter import num_arg
from liquid.filter import int_arg

DecimalT = decimal.Decimal
NumberT = Union[float, int]


@math_filter
def abs_(num: NumberT) -> NumberT:
    """Return the absolute value of a number.

    Accepts an int, float or a string representations of an int or float.
    """
    return abs(num)


@math_filter
def at_most(num: NumberT, other: NumberT) -> NumberT:
    """Return `val` or `args[0]`, whichever is smaller.

    Accepts an int, float or a string representations of an int or float.
    """
    other = num_arg(other, default=0)
    return min(num, other)


@math_filter
def at_least(num: NumberT, other: NumberT) -> NumberT:
    """Return `val` or `args[0]`, whichever is greater.

    Accepts an int, float or a string representations of an int or float.
    """
    other = num_arg(other, default=0)
    return max(num, other)


@math_filter
def ceil(num: NumberT) -> NumberT:
    """Return the ceiling of x as an Integral.

    Accepts an int, float or a string representations of an int or float.
    """
    return math.ceil(num)


@math_filter
def divided_by(num: NumberT, other: object) -> NumberT:
    """Divide `num` by `other`."""
    other = num_arg(other, default=0)

    try:
        if isinstance(other, int) and isinstance(num, int):
            return num // other
        return num / other
    except ZeroDivisionError as err:
        raise FilterArgumentError(f"divided_by: can't divide by {other}") from err


@math_filter
def floor(num: NumberT) -> NumberT:
    """Return the floor of x as an Integral.

    Accepts an int, float or a string representations of an int or float.
    """
    return math.floor(num)


@math_filter
def minus(num: NumberT, other: NumberT) -> NumberT:
    """Subtract one number from another."""
    other = num_arg(other, default=0)

    if isinstance(num, int) and isinstance(other, int):
        return num - other
    return float(DecimalT(str(num)) - DecimalT(str(other)))


@math_filter
def plus(num: NumberT, other: NumberT) -> NumberT:
    """Add one number to another."""
    other = num_arg(other, default=0)

    if isinstance(num, int) and isinstance(other, int):
        return num + other
    return float(DecimalT(str(num)) + DecimalT(str(other)))


@math_filter
def round_(num: NumberT, ndigits: Optional[int] = None) -> NumberT:
    """Round a number to a given precision in decimal digits."""
    if ndigits or is_undefined(ndigits):
        ndigits = int_arg(ndigits, default=0)
        return round(num, ndigits)
    return round(num)


@math_filter
def times(num: NumberT, other: NumberT) -> NumberT:
    """Multiply a value by an integer or float."""
    other = num_arg(other, default=0)

    if isinstance(num, int) and isinstance(other, int):
        return num * other
    return float(DecimalT(str(num)) * DecimalT(str(other)))


@math_filter
def modulo(num: NumberT, other: NumberT) -> NumberT:
    """Divide a value by a number and returns the remainder."""
    other = num_arg(other, default=0)

    try:
        if isinstance(num, int) and isinstance(other, int):
            return num % other
        return float(DecimalT(str(num)) % DecimalT(str(other)))

    except ZeroDivisionError as err:
        raise FilterArgumentError(f"modulo: can't divide by {other}") from err



"""Template loader test cases."""
# pylint: disable=missing-class-docstring
import asyncio
import tempfile
import time
import unittest

from pathlib import Path

from typing import Dict

from liquid import Environment
from liquid import Context
from liquid.template import BoundTemplate
from liquid.template import AwareBoundTemplate

from liquid.loaders import BaseLoader
from liquid.loaders import FileSystemLoader
from liquid.loaders import FileExtensionLoader
from liquid.loaders import ChoiceLoader
from liquid.loaders import DictLoader
from liquid.loaders import TemplateSource

from liquid.exceptions import TemplateNotFound


class FileSystemLoaderTestCase(unittest.TestCase):
    """Test loading templates from the file system."""

    def test_load_template(self):
        """Test that we can load a template from the file system."""
        env = Environment(loader=FileSystemLoader(search_path="tests/fixtures/"))
        template = env.get_template(name="dropify/index.liquid")
        self.assertIsInstance(template, BoundTemplate)

    def test_cached_template(self):
        """Test that templates loaded from the file system get cached."""
        env = Environment(
            loader=FileSystemLoader(search_path="tests/fixtures/"),
            auto_reload=True,
        )
        template = env.get_template(name="dropify/index.liquid")
        self.assertTrue(template.is_up_to_date)

        another = env.get_template(name="dropify/index.liquid")
        self.assertTrue(another.is_up_to_date)

        self.assertEqual(template.tree, another.tree)

    def test_auto_reload_template(self):
        """Test templates loaded from the file system are reloaded automatically."""
        with tempfile.TemporaryDirectory() as tmpdir:
            template_path = Path(tmpdir) / "somefile.txt"

            # Initial template content
            with template_path.open("w", encoding="UTF-8") as fd:
                fd.write("hello there\n")

            env = Environment(
                loader=FileSystemLoader(search_path=tmpdir),
                auto_reload=True,
            )

            template = env.get_template(name=str(template_path))
            self.assertTrue(template.is_up_to_date)

            same_template = env.get_template(name=str(template_path))
            self.assertTrue(template.is_up_to_date)

            self.assertEqual(template.tree, same_template.tree)

            # Update template content.
            time.sleep(0.01)  # Make sure some time has passed.
            template_path.touch()

            # Template has been updated
            self.assertFalse(template.is_up_to_date)

            updated_template = env.get_template(name=str(template_path))
            self.assertTrue(updated_template.is_up_to_date)

            self.assertNotEqual(template.tree, updated_template.tree)

    def test_without_auto_reload_template(self):
        """Test that auto_reload can be disabled."""
        with tempfile.TemporaryDirectory() as tmpdir:
            template_path = Path(tmpdir) / "somefile.txt"

            # Initial template content
            with template_path.open("w", encoding="UTF-8") as fd:
                fd.write("hello there\n")

            env = Environment(
                loader=FileSystemLoader(search_path=tmpdir),
                auto_reload=False,
            )

            template = env.get_template(name=str(template_path))
            self.assertTrue(template.is_up_to_date)

            same_template = env.get_template(name=str(template_path))
            self.assertTrue(template.is_up_to_date)

            self.assertEqual(template.tree, same_template.tree)

            # Update template content.
            time.sleep(0.01)  # Make sure some time has passed.
            template_path.touch()

            # Template has been updated
            self.assertFalse(template.is_up_to_date)

            updated_template = env.get_template(name=str(template_path))
            self.assertFalse(updated_template.is_up_to_date)

            self.assertEqual(template.tree, updated_template.tree)

    def test_template_cache_size(self):
        """Test that we can control the template cache size."""
        with tempfile.TemporaryDirectory() as tmpdir:
            template_path = Path(tmpdir) / "somefile.txt"
            another_template_path = Path(tmpdir) / "otherfile.txt"

            # Initial template content
            with template_path.open("w", encoding="UTF-8") as fd:
                fd.write("hello there\n")

            with another_template_path.open("w", encoding="UTF-8") as fd:
                fd.write("goodbye there\n")

            # Cache size of zero sets auto_reload to False
            env = Environment(
                loader=FileSystemLoader(search_path=tmpdir),
                cache_size=0,
            )
            self.assertFalse(env.auto_reload)

            # Very small cache size.
            env = Environment(
                loader=FileSystemLoader(search_path=tmpdir),
                cache_size=1,
            )

            template = env.get_template(name=str(template_path))
            self.assertTrue(template.is_up_to_date)

            same_template = env.get_template(name=str(template_path))
            self.assertTrue(template.is_up_to_date)

            # Cached tempalate is the same object
            self.assertEqual(template.tree, same_template.tree)

            # Push the first template out of the cache
            another_template = env.get_template(name=str(another_template_path))
            self.assertTrue(another_template.is_up_to_date)
            self.assertEqual(len(env.cache), 1)

    def test_template_not_found(self):
        """Test that we get an error if the template does not exist."""
        env = Environment(loader=FileSystemLoader(search_path="tests/fixtures/"))
        with self.assertRaises(TemplateNotFound):
            env.get_template(name="dropify/nosuchthing.liquid")

    def test_no_such_search_path(self):
        """Test that a non-existant search path does not cause an error."""
        env = Environment(loader=FileSystemLoader(search_path="nosuchthing/foo/"))
        with self.assertRaises(TemplateNotFound):
            env.get_template(name="dropify/nosuchthing.liquid")

    def test_multiple_search_paths(self):
        """Test that we can search multiple directories for templates."""
        env = Environment(
            loader=FileSystemLoader(
                search_path=[
                    "tests/fixtures/",
                    "tests/fixtures/subfolder/",
                ]
            )
        )

        template = env.get_template(name="fallback.html")
        self.assertIsInstance(template, BoundTemplate)
        self.assertEqual(template.path, Path("tests/fixtures/subfolder/fallback.html"))

    def test_stay_in_search_path(self):
        """Test that we can't stray above the search path"""
        env = Environment(
            loader=FileSystemLoader(search_path="tests/fixtures/subfolder")
        )

        with self.assertRaises(TemplateNotFound):
            _ = env.get_template(name="../dropify/index.liquid")


class TemplateDropTestCase(unittest.TestCase):
    def setUp(self):
        self.env = Environment(
            loader=DictLoader(
                {
                    "somename": "{{ template.name }}",
                    "somedir/somename.liquid": "{{ template.directory }}",
                    "somename.somesuffix.liquid": "{{ template.suffix }}",
                }
            )
        )

        self.env.template_class = AwareBoundTemplate

    def test_template_name(self):
        """Test that templates have access to their name."""
        template = self.env.get_template(name="somename")
        self.assertEqual(template.render(), "somename")

    def test_template_directory(self):
        """Test that templates have access to their directory name."""
        template = self.env.get_template(name="somedir/somename.liquid")
        self.assertEqual(template.render(), "somedir")

    def test_template_suffix(self):
        """Test that templates have access to their suffix."""
        template = self.env.get_template(name="somename.somesuffix.liquid")
        self.assertEqual(template.render(), "somesuffix")

    def test_drop_contains(self):
        """Test that we can check drop membership."""
        template = self.env.get_template(name="somename")
        assert isinstance(template, AwareBoundTemplate)
        self.assertEqual("name" in template.drop, True)
        self.assertEqual("foo" in template.drop, False)

    def test_drop_length(self):
        """Test that we get the length of a template drop"""
        template = self.env.get_template(name="somename")
        assert isinstance(template, AwareBoundTemplate)
        self.assertEqual(len(template.drop), 3)

    def test_iter_drop(self):
        """Test that we can iterate a template drop."""
        template = self.env.get_template(name="somename")
        assert isinstance(template, AwareBoundTemplate)
        keys = list(template.drop)
        self.assertEqual(keys, ["directory", "name", "suffix"])


class MatterDictLoader(DictLoader):
    def __init__(
        self,
        templates: Dict[str, str],
        matter: Dict[str, Dict[str, object]],
    ):
        super().__init__(templates)
        self.matter = matter

    def get_source(self, _: Environment, template_name: str) -> TemplateSource:
        try:
            source = self.templates[template_name]
        except KeyError as err:
            raise TemplateNotFound(template_name) from err

        return TemplateSource(
            source=source,
            filename=template_name,
            uptodate=None,
            matter=self.matter.get(template_name),
        )


class MatterLoaderTestCase(unittest.TestCase):
    def test_matter_loader(self):
        """Test that template loaders can add to render context."""
        loader = MatterDictLoader(
            templates={
                "some": "Hello, {{ you }}{{ username }}!",
                "other": "Goodbye, {{ you }}{{ username }}.",
                "thing": "{{ you }}{{ username }}",
            },
            matter={
                "some": {"you": "World"},
                "other": {"username": "Smith"},
            },
        )

        env = Environment(loader=loader)

        template = env.get_template("some")
        self.assertEqual(template.render(), "Hello, World!")

        template = env.get_template("other")
        self.assertEqual(template.render(), "Goodbye, Smith.")

        template = env.get_template("thing")
        self.assertEqual(template.render(), "")

    def test_matter_global_priority(self):
        """Test that matter variables take priority over globals."""
        loader = MatterDictLoader(
            templates={"some": "Hello, {{ you }}!"},
            matter={"some": {"you": "Liquid"}},
        )

        env = Environment(loader=loader, globals={"you": "World"})
        template = env.get_template("some", globals={"you": "Jinja"})

        self.assertEqual(template.render(), "Hello, Liquid!")

    def test_matter_local_priority(self):
        """Test that render args take priority over matter variables."""
        loader = MatterDictLoader(
            templates={"some": "Hello, {{ you }}!"},
            matter={"some": {"you": "Liquid"}},
        )

        env = Environment(loader=loader)
        template = env.get_template("some")

        self.assertEqual(template.render(you="John"), "Hello, John!")


class ChoiceLoaderTestCase(unittest.TestCase):
    """Test loading templates from a list of loaders."""

    def test_choose_between_loaders(self):
        """Test that we can load templates from a list of loaders."""
        loader = ChoiceLoader(
            loaders=[
                DictLoader({"a": "Hello, {{ you }}!"}),
                DictLoader(
                    {
                        "a": "unreachable",
                        "b": "the quick brown {{ animal | default: 'fox' }}",
                    }
                ),
            ]
        )

        env = Environment(loader=loader)

        template = env.get_template("a")
        self.assertEqual(template.render(you="World"), "Hello, World!")

        template = env.get_template("b")
        self.assertEqual(template.render(), "the quick brown fox")

        with self.assertRaises(TemplateNotFound):
            env.get_template("c")


class FileExtensionLoaderTestCase(unittest.TestCase):
    """Test loading templates from the file system with automatic extensions."""

    def test_load_template_with_missing_extension(self):
        """Test that we can load a template from the file system when a file extension
        is missing."""
        env = Environment(
            loader=FileExtensionLoader(
                search_path="tests/fixtures/",
                ext=".liquid",
            )
        )
        template = env.get_template(name="dropify/index")
        self.assertIsInstance(template, BoundTemplate)

    def test_stay_in_search_path(self):
        """Test that we can't stray above the search path"""
        env = Environment(
            loader=FileExtensionLoader(
                search_path="tests/fixtures/subfolder",
                ext=".liquid",
            )
        )

        with self.assertRaises(TemplateNotFound):
            _ = env.get_template(name="../dropify/index")

    def test_multiple_search_paths(self):
        """Test that we can search multiple directories for templates."""
        env = Environment(
            loader=FileExtensionLoader(
                search_path=[
                    "tests/fixtures/",
                    "tests/fixtures/subfolder/",
                ],
                ext=".liquid",
            )
        )

        template = env.get_template(name="fallback.html")
        self.assertIsInstance(template, BoundTemplate)
        self.assertEqual(template.path, Path("tests/fixtures/subfolder/fallback.html"))

    def test_template_not_found(self):
        """Test that we get an error if the template does not exist."""
        env = Environment(
            loader=FileExtensionLoader(
                search_path="tests/fixtures/",
                ext=".liquid",
            )
        )
        with self.assertRaises(TemplateNotFound):
            env.get_template(name="dropify/nosuchthing")


class BadLoader(BaseLoader):
    pass


class BaseLoaderTestCase(unittest.TestCase):
    """Test case for the abstract base loader."""

    def test_get_source_is_required(self):
        """Test that the `get_source` method is required."""
        env = Environment(loader=BadLoader())

        with self.assertRaises(TemplateNotFound):
            env.get_template(name="somename")


class MockContextLoader(DictLoader):
    def __init__(self, templates: Dict[str, str]):
        self.kwargs = {}
        super().__init__(templates)

    def get_source_with_context(
        self, context: Context, template_name: str, **kwargs: str
    ) -> TemplateSource:
        self.kwargs.update(kwargs)
        self.kwargs["uid"] = context.resolve("uid", default=None)
        return super().get_source(context.env, template_name)

    async def get_source_with_context_async(
        self, context: Context, template_name: str, **kwargs: str
    ) -> TemplateSource:
        self.kwargs.update(kwargs)
        self.kwargs["uid"] = context.resolve("uid", default=None)
        return await super().get_source_async(context.env, template_name)


class ContextLoaderTestCase(unittest.TestCase):
    """Test case for a loader that used context."""

    def test_keyword_arguments(self):
        """Test that keyword arguments passed to `get_template_with_context` are
        available to a context loader."""
        loader = MockContextLoader({"snippet": "hello"})
        env = Environment(loader=loader, cache_size=0)
        template = env.from_string("{% include 'snippet' %}")

        template.render()
        self.assertIn("tag", loader.kwargs)
        self.assertEqual(loader.kwargs["tag"], "include")

        # and async
        loader.kwargs.clear()
        self.assertNotIn("tag", loader.kwargs)

        async def coro():
            return await template.render_async()

        asyncio.run(coro())
        self.assertIn("tag", loader.kwargs)
        self.assertEqual(loader.kwargs["tag"], "include")

    def test_resolve_from_context(self):
        """Test context loaders can resolve render context variables."""
        loader = MockContextLoader({"snippet": "hello"})
        env = Environment(loader=loader, cache_size=0)
        template = env.from_string("{% include 'snippet' %}", globals={"uid": 1234})

        template.render()
        self.assertIn("uid", loader.kwargs)
        self.assertEqual(loader.kwargs["uid"], 1234)

        # and async
        loader.kwargs.clear()
        self.assertNotIn("uid", loader.kwargs)

        async def coro():
            return await template.render_async()

        asyncio.run(coro())
        self.assertIn("uid", loader.kwargs)
        self.assertEqual(loader.kwargs["uid"], 1234)



"""Functions for parsing filtered expressions. Like those found in output statements and
assign tags.
"""
from itertools import islice

from typing import Dict
from typing import List
from typing import Iterator
from typing import Iterable
from typing import Tuple

from liquid.expression import Expression
from liquid.expression import Filter
from liquid.expression import FilteredExpression

from liquid.expressions.common import Token
from liquid.expressions.common import parse_blank
from liquid.expressions.common import parse_boolean
from liquid.expressions.common import parse_empty
from liquid.expressions.common import parse_float_literal
from liquid.expressions.common import parse_identifier
from liquid.expressions.common import parse_integer_literal
from liquid.expressions.common import parse_nil
from liquid.expressions.common import parse_string_literal
from liquid.expressions.common import make_parse_range

from liquid.expressions.filtered.lex import tokenize
from liquid.expressions.stream import TokenStream

from liquid.exceptions import LiquidSyntaxError

from liquid.token import TOKEN_EOF
from liquid.token import TOKEN_PIPE
from liquid.token import TOKEN_COMMA
from liquid.token import TOKEN_FALSE
from liquid.token import TOKEN_TRUE
from liquid.token import TOKEN_NIL
from liquid.token import TOKEN_NULL
from liquid.token import TOKEN_EMPTY
from liquid.token import TOKEN_BLANK
from liquid.token import TOKEN_STRING
from liquid.token import TOKEN_INTEGER
from liquid.token import TOKEN_FLOAT
from liquid.token import TOKEN_IDENTIFIER
from liquid.token import TOKEN_LPAREN
from liquid.token import TOKEN_COLON
from liquid.token import TOKEN_RANGE_LITERAL


TOKEN_MAP = {
    TOKEN_FALSE: parse_boolean,
    TOKEN_TRUE: parse_boolean,
    TOKEN_NIL: parse_nil,
    TOKEN_NULL: parse_nil,
    TOKEN_EMPTY: parse_empty,
    TOKEN_BLANK: parse_blank,
    TOKEN_STRING: parse_string_literal,
    TOKEN_INTEGER: parse_integer_literal,
    TOKEN_FLOAT: parse_float_literal,
    TOKEN_IDENTIFIER: parse_identifier,
}


def parse_obj(stream: TokenStream) -> Expression:
    """Parse an object from the stream of tokens.

    An object could be a constant, like `true` or `nil`, a range literal or an
    identifier. An identifier could be chained, possibly with nested identifiers
    between square brackets.
    """
    try:
        return TOKEN_MAP[stream.current[1]](stream)
    except KeyError as err:
        raise LiquidSyntaxError(
            f"unexpected {stream.current[2]!r}", linenum=stream.current[0]
        ) from err


parse_range = make_parse_range(parse_obj)


def parse_range_with_parens(stream: TokenStream) -> Expression:
    """Like `parse_range` but consumes the extra `RANGE_LPAREN` token first."""
    stream.expect(TOKEN_RANGE_LITERAL)
    next(stream)  # Eat extra token
    return parse_range(stream)


TOKEN_MAP[TOKEN_LPAREN] = parse_range
TOKEN_MAP[TOKEN_RANGE_LITERAL] = parse_range_with_parens


def split_at_pipe(tokens: Iterable[Token]) -> Iterator[List[Token]]:
    """Split tokens on into lists, using TOKEN_PIPE as the delimiter."""
    buf: List[Token] = []
    for token in tokens:
        if token[1] == TOKEN_PIPE:
            yield buf
            buf = []
        else:
            buf.append(token)
    yield buf


def split_at_first_pipe(tokens: Iterable[Token]) -> Iterator[List[Token]]:
    """Like split_at_pipe, but stops splitting after the first TOKEN_PIPE is found."""
    buf: List[Token] = []
    for token in tokens:
        if token[1] == TOKEN_PIPE:
            yield buf
            yield list(tokens)
            return
        buf.append(token)
    yield buf


def split_at_comma(
    tokens: Iterable[Token],
) -> Iterator[List[Token]]:  # pragma: no cover
    """Split tokens into lists, using TOKEN_COMMA as the delimiter."""
    buf: List[Token] = []
    for token in tokens:
        if token[1] == TOKEN_COMMA:
            yield buf
            buf = []
        else:
            buf.append(token)
    yield buf


def bucket_args(
    arguments: Iterable[Tuple[str, Expression]]
) -> Tuple[List[Expression], Dict[str, Expression]]:  # pragma: no cover
    """Split filter arguments into positional and keyword arguments."""
    args = []
    kwargs = {}
    for name, expr in arguments:
        if not name:
            args.append(expr)
        else:
            kwargs[name] = expr
    return args, kwargs


def parse_filter(tokens: List[Token], linenum: int = 1) -> Filter:  # pragma: no cover
    """Parse a Liquid filter from a list of tokens."""
    if not tokens:
        raise LiquidSyntaxError(
            "unexpected pipe or missing filter name", linenum=linenum
        )

    name = tokens[0][2]

    if len(tokens) > 1:
        if tokens[1][1] != TOKEN_COLON:
            raise LiquidSyntaxError(
                f"expected a colon after {name!r}",
                linenum=tokens[1][0],
            )
        return Filter(name, *bucket_args(parse_args(islice(tokens, 2, None))))
    return Filter(name, [])


def parse_args(
    tokens: Iterator[Token],
) -> Iterator[Tuple[str, Expression]]:  # pragma: no cover
    """Parse a filter's arguments from the given token iterator."""
    for arg_tokens in split_at_comma(tokens):
        yield parse_arg(arg_tokens)


def parse_arg(tokens: List[Token]) -> Tuple[str, Expression]:  # pragma: no cover
    """Parse a single argument from a list of tokens."""
    if len(tokens) > 1 and tokens[1][1] == TOKEN_COLON:
        # A named/keyword parameter/argument
        return tokens[0][2], parse_obj(TokenStream(islice(tokens, 2, None)))
    return "", parse_obj(TokenStream(iter(tokens)))


# _parse_filter has replaced the parse_* functions above in the name of better syntax
# error handling. On the off chance someone is using them, the parse_* functions will
# be depreciated as we approach Python Liquid version 2 and then removed.


def _parse_filter(tokens: List[Token], linenum: int) -> Filter:
    if not tokens:
        raise LiquidSyntaxError(
            "unexpected pipe or missing filter name", linenum=linenum
        )

    stream = TokenStream(iter(tokens))
    stream.expect(TOKEN_IDENTIFIER)
    name = stream.current[2]

    next(stream)
    # Shortcut for filters with no arguments.
    if stream.current[1] == TOKEN_EOF:
        return Filter(name, [])

    # Eat colon
    stream.expect(TOKEN_COLON)
    next(stream)

    args: List[Expression] = []
    kwargs: Dict[str, Expression] = {}

    while stream.current[1] != TOKEN_EOF:
        if stream.peek[1] == TOKEN_COLON:
            # A keyword argument
            stream.expect(TOKEN_IDENTIFIER)
            key = next(stream)[2]
            # Eat colon
            next(stream)
            kwargs[key] = parse_obj(stream)
        else:
            # A positional argument
            args.append(parse_obj(stream))

        # Eat comma
        next(stream)
        if stream.current[1] != TOKEN_EOF:
            stream.expect(TOKEN_COMMA)
            next(stream)

    return Filter(name, args, kwargs)


def parse_from_tokens(tokens: Iterator[Token], linenum: int = 1) -> FilteredExpression:
    """Parse an expression with zero or more filters from a token iterator."""
    parts = tuple(split_at_first_pipe(tokens))
    if len(parts) == 1:
        stream = TokenStream(iter(parts[0]))
        return FilteredExpression(parse_obj(stream))

    left = parse_obj(TokenStream(iter(parts[0])))
    filters = [_parse_filter(_tokens, linenum) for _tokens in split_at_pipe(parts[1])]
    return FilteredExpression(left, filters)


def parse(expr: str, linenum: int = 1) -> FilteredExpression:
    """Parse an expression string with zero or more filters."""
    return parse_from_tokens(tokenize(expr, linenum))



"""Golden tests cases for testing liquid's built-in `uniq` filter."""

from liquid.golden.case import Case

cases = [
    Case(
        description="array of strings",
        template=r"{{ a | uniq | join: '#' }}",
        expect="a#b",
        globals={"a": ["a", "b", "b", "a"]},
    ),
    Case(
        description="array of things",
        template=r"{{ a | uniq | join: '#' }}",
        expect="a#b#1",
        globals={"a": ["a", "b", 1, 1]},
    ),
    Case(
        description="empty array",
        template=r"{{ a | uniq | join: '#' }}",
        expect="",
        globals={"a": []},
    ),
    Case(
        description="unhashable items",
        template=r"{{ a | uniq | join: '#' }}",
        expect=r"a#b#{}",
        globals={"a": ["a", "b", [], {}, {}]},
    ),
    Case(
        description="left value is not an array",
        template=r"{{ a | uniq | join: '#' }}",
        expect="123",
        globals={"a": 123},
    ),
    Case(
        description="left value is undefined",
        template=r"{{ nosuchthing | uniq | join: '#' }}",
        expect="",
    ),
    Case(
        description="too many arguments",
        template=r"{{ nosuchthing | uniq: 'foo', 'bar' }}",
        expect="",
        error=True,
    ),
    Case(
        description="array of objects with key property",
        template=(
            r"{% assign x = a | uniq: 'title' %}"
            r"{% for obj in x %}"
            r"{% for i in obj %}"
            r"({{ i[0] }},{{ i[1] }})"
            r"{% endfor %}"
            r"{% endfor %}"
        ),
        expect="(title,foo)(name,a)(title,bar)(name,c)",
        globals={
            "a": [
                {"title": "foo", "name": "a"},
                {"title": "foo", "name": "b"},
                {"title": "bar", "name": "c"},
            ]
        },
    ),
    Case(
        description="array of objects with missing key property",
        template=(
            r"{% assign x = a | uniq: 'title' %}"
            r"{% for obj in x %}"
            r"{% for i in obj %}"
            r"({{ i[0] }},{{ i[1] }})"
            r"{% endfor %}"
            r"{% endfor %}"
        ),
        expect="(title,foo)(name,a)(title,bar)(name,c)(heading,bar)(name,c)",
        globals={
            "a": [
                {"title": "foo", "name": "a"},
                {"title": "foo", "name": "b"},
                {"title": "bar", "name": "c"},
                {"heading": "bar", "name": "c"},
                {"heading": "baz", "name": "d"},
            ]
        },
    ),
]



"""Filter functions that operate on strings."""
from __future__ import annotations

import base64
import binascii
import html
import re
import urllib.parse

from typing import Any
from typing import List
from typing import Optional
from typing import Union
from typing import TYPE_CHECKING

try:
    from markupsafe import escape as markupsafe_escape
    from markupsafe import Markup
    from markupsafe import soft_str
except ImportError:
    from liquid.exceptions import escape as markupsafe_escape  # type: ignore
    from liquid.exceptions import Markup  # type: ignore

    # pylint: disable=invalid-name
    soft_str = str  # type: ignore

from liquid import is_undefined
from liquid.exceptions import FilterArgumentError
from liquid.exceptions import FilterError

from liquid.filter import with_environment
from liquid.filter import string_filter
from liquid.filter import liquid_filter

from liquid.limits import to_int

from liquid.utils.html import strip_tags
from liquid.utils.text import truncate_chars

if TYPE_CHECKING:
    from liquid import Environment


@string_filter
def append(val: str, arg: object) -> str:
    """Concatenate two strings."""
    if not isinstance(arg, str):
        arg = str(arg)
    return val + arg


@string_filter
def capitalize(val: str) -> str:
    """Make sure the first character of a string is upper case and the rest
    lowercase."""
    return val.capitalize()


@string_filter
def downcase(val: str) -> str:
    """Make all characters in a string lower case."""
    return val.lower()


@with_environment
@string_filter
def escape(val: str, *, environment: Environment) -> str:
    """Convert the characters &, < and > in string s to HTML-safe sequences."""
    if environment.autoescape:
        return markupsafe_escape(str(val))
    return html.escape(val)


@with_environment
@string_filter
def escape_once(val: str, *, environment: Environment) -> str:
    """Convert the characters &, < and > in string s to HTML-safe sequences."""
    if environment.autoescape:
        return Markup(val).unescape()
    return html.escape(html.unescape(val))


@string_filter
def lstrip(val: str) -> str:
    """Remove leading whitespace."""
    return val.lstrip()


RE_LINETERM = re.compile(r"\r?\n")


@with_environment
@string_filter
def newline_to_br(val: str, *, environment: Environment) -> str:
    """Convert '\\r\\n' or '\\n' to '<br />\\n'."""
    # The reference implementation was changed to replace "\r\n" as well as "\n",
    # but they don't seem to care about "\r" (Mac OS).
    if environment.autoescape:
        val = markupsafe_escape(val)
        return Markup(RE_LINETERM.sub("<br />\n", val))
    return RE_LINETERM.sub("<br />\n", val)


@string_filter
def prepend(val: str, arg: str) -> str:
    """Concatenate string value to argument string."""
    return soft_str(arg) + val


@string_filter
def remove(val: str, arg: str) -> str:
    """Remove all occurrences of argument string from value."""
    return val.replace(soft_str(arg), "")


@string_filter
def remove_first(val: str, arg: str) -> str:
    """Remove the first occurrence of the argument string from value."""
    return val.replace(soft_str(arg), "", 1)


@string_filter
def remove_last(val: str, arg: str) -> str:
    """Remove the last occurrence of the argument string from value."""
    try:
        before, _, after = val.rpartition(soft_str(arg))
    except ValueError:
        # empty separator
        return val
    if before:
        return before + after
    return val


@string_filter
def replace(val: str, seq: str, sub: str = "") -> str:
    """Replaces every occurrence of the first argument in a string with the second
    argument."""
    return val.replace(soft_str(seq), soft_str(sub))


@string_filter
def replace_first(val: str, seq: str, sub: str = "") -> str:
    """Replaces the first occurrence of the first argument in a string with the second
    argument."""
    return val.replace(soft_str(seq), soft_str(sub), 1)


@string_filter
def replace_last(val: str, seq: str, sub: str) -> str:
    """Replace the last occurrence of the first argument in a string with the second
    argument."""
    try:
        before, _, after = val.rpartition(soft_str(seq))
    except ValueError:
        # empty separator
        return val + soft_str(sub)
    if before:
        return before + soft_str(sub) + after
    return val


@string_filter
def upcase(val: str) -> str:
    """Make all characters in a string upper case."""
    return val.upper()


MAX_SLICE_ARG = (1 << 63) - 1
MIN_SLICE_ARG = -(1 << 63)


def _slice_arg(val: Any) -> int:
    # The reference implementation does not cast floats to int.
    if isinstance(val, float):
        raise FilterArgumentError(
            f"slice expected an integer start, found {type(val).__name__}"
        )

    try:
        rv = to_int(val)
    except (ValueError, TypeError) as err:
        raise FilterArgumentError(
            f"slice expected an integer start, found {type(val).__name__}"
        ) from err

    rv = min(rv, MAX_SLICE_ARG)
    rv = max(rv, MIN_SLICE_ARG)
    return rv


@liquid_filter
def slice_(val: Any, start: Any, length: Any = 1) -> Union[str, List[object]]:
    """Return a substring or subsequence, starting at `start`, containing up to
    `length` characters or items.

    Array-like objects return a list, strings return a substring, all other objects are
    cast to a string before returning a substring.
    """
    if not isinstance(val, (list, tuple, range, str)):
        val = str(val)

    if is_undefined(start):
        raise FilterArgumentError("slice expected an integer, found Undefined")

    if is_undefined(length):
        length = 1

    _start = _slice_arg(start)
    _length = _slice_arg(length)
    end: Optional[int] = _start + _length

    # A negative start index and a length that exceeds the theoretical length
    # of the sequence.
    if isinstance(end, int) and _start < 0 <= end:
        end = None

    if isinstance(val, str):
        return val[_start:end]
    return list(val[_start:end])


@string_filter
def split(val: str, seq: str) -> List[str]:
    """Split a string into a list of string, using the argument as a delimiter."""
    if not seq:
        return list(val)

    return val.split(soft_str(seq))


@string_filter
def strip(val: str) -> str:
    """Remove leading and trailing whitespace."""
    return val.strip()


@string_filter
def rstrip(val: str) -> str:
    """Remove trailing whitespace."""
    return val.rstrip()


@with_environment
@string_filter
def strip_html(val: str, *, environment: Environment) -> str:
    """Return the given HTML with all HTML tags removed."""
    stripped = strip_tags(val)
    if environment.autoescape and isinstance(val, Markup):
        return Markup(stripped)
    return stripped


@with_environment
@string_filter
def strip_newlines(val: str, *, environment: Environment) -> str:
    """Return the given string with all newline characters removed."""
    if environment.autoescape:
        val = markupsafe_escape(val)
        return Markup(RE_LINETERM.sub("", val))
    return RE_LINETERM.sub("", val)


@string_filter
def truncate(val: str, num: Any = 50, end: str = "...") -> str:
    """Truncate a string if it is longer than the specified number of characters."""
    if is_undefined(num):
        raise FilterArgumentError("truncate expected an integer, found Undefined")

    try:
        num = to_int(num)
    except ValueError as err:
        raise FilterArgumentError(
            f"truncate expected an integer, found {type(num).__name__}"
        ) from err

    end = str(end)
    return truncate_chars(val, num, end)


# Mimic reference implementation's limit to the number of words that can be truncated.
MAX_TRUNC_WORDS = (1 << 31) - 1


@string_filter
def truncatewords(val: str, num: Any = 15, end: str = "...") -> str:
    """Shorten a string down to the given number of words."""
    if is_undefined(num):
        raise FilterArgumentError("truncate expected an integer, found Undefined")

    try:
        num = to_int(num)
    except ValueError as err:
        raise FilterArgumentError(
            f"truncate expected an integer, found {type(num).__name__}"
        ) from err

    end = str(end)

    # The reference implementation seems to force a minimum `num` of 1.
    if num <= 0:
        num = 1

    # Replaces consecutive whitespace with a single newline.
    words = val.split()

    # This too mimics the reference implementation's big integer work around.
    if num >= MAX_TRUNC_WORDS:
        return val

    if len(words) < num:
        return " ".join(words)

    return " ".join(words[:num]) + end


@with_environment
@string_filter
def url_encode(val: str, *, environment: Environment) -> str:
    """Percent encode a string so it is useable in a URL."""
    if environment.autoescape:
        return Markup(urllib.parse.quote_plus(val))
    return urllib.parse.quote_plus(val)


@string_filter
def url_decode(val: str) -> str:
    """Decode a string that has been URL encoded."""
    # Assuming URL decoded strings are all unsafe.
    return urllib.parse.unquote_plus(val)


@string_filter
def base64_encode(val: str) -> str:
    """Encode a string to base64."""
    return base64.b64encode(val.encode()).decode()


@string_filter
def base64_decode(val: str) -> str:
    """Decode a string from base64.

    The decoded value is assumed to be UTF-8 and will be decoded as UTF-8.
    """
    try:
        return base64.b64decode(val).decode()
    except binascii.Error as err:
        raise FilterError("invalid base64-encoded string") from err


@string_filter
def base64_url_safe_encode(val: str) -> str:
    """Encode a string to URL safe base64."""
    return base64.urlsafe_b64encode(val.encode()).decode()


@string_filter
def base64_url_safe_decode(val: str) -> str:
    """Decode a URL safe string from base64.

    The decoded value is assumed to be UTF-8 and will be decoded as UTF-8.
    """
    try:
        return base64.urlsafe_b64decode(val).decode()
    except binascii.Error as err:
        raise FilterError("invalid base64-encoded string") from err



"""Liquid template render context."""

from __future__ import annotations

import collections.abc
import datetime
import itertools
import re
import sys
import warnings

from collections import deque
from contextlib import contextmanager

from functools import partial
from functools import reduce

from itertools import cycle
from io import StringIO

from operator import getitem
from operator import mul

from typing import Any
from typing import Callable
from typing import Dict
from typing import Iterable
from typing import Iterator
from typing import List
from typing import Mapping
from typing import Optional
from typing import Sequence
from typing import TextIO
from typing import Union
from typing import TYPE_CHECKING

from liquid.exceptions import NoSuchFilterFunc
from liquid.exceptions import ContextDepthError
from liquid.exceptions import Error
from liquid.exceptions import LocalNamespaceLimitError
from liquid.exceptions import LoopIterationLimitError
from liquid.exceptions import UndefinedError
from liquid.exceptions import lookup_warning

from liquid import Mode
from liquid.output import LimitedStringIO

if TYPE_CHECKING:  # pragma: no cover
    from liquid import Environment
    from liquid.template import BoundTemplate
    from liquid.builtin.tags.for_tag import ForLoop


ContextPath = Union[str, Sequence[Union[str, int]]]
Namespace = Mapping[str, object]

_undefined = object()


# pylint: disable=too-many-return-statements
def _getitem(obj: Any, key: Any) -> Any:
    """Item getter with special methods for arrays/lists and hashes/dicts."""
    # NOTE: A runtime checkable protocol was too slow.
    if hasattr(key, "__liquid__"):
        key = key.__liquid__()

    if key == "size":
        try:
            return getitem(obj, "size")
        except (KeyError, IndexError, TypeError):
            if isinstance(obj, collections.abc.Sized):
                return len(obj)
            raise

    if key == "first" and not isinstance(obj, str):
        try:
            return getitem(obj, "first")
        except (KeyError, IndexError, TypeError):
            if isinstance(obj, collections.abc.Mapping) and obj:
                return list(itertools.islice(obj.items(), 1))[0]
            if isinstance(obj, collections.abc.Sequence):
                return obj[0]
            raise

    if key == "last" and not isinstance(obj, str):
        try:
            return getitem(obj, "last")
        except (KeyError, IndexError, TypeError):
            if isinstance(obj, collections.abc.Sequence):
                return obj[-1]
            raise

    return getitem(obj, key)


# pylint: disable=too-many-return-statements
async def _getitem_async(obj: Any, key: Any) -> object:
    """Item getter with special methods for arrays/lists and hashes/dicts."""
    # NOTE: A runtime checkable protocol was too slow.
    if hasattr(key, "__liquid__"):
        key = key.__liquid__()

    if key == "size":
        try:
            return (
                await obj.__getitem_async__(key)
                if hasattr(obj, "__getitem_async__")
                else getitem(obj, "size")
            )
        except (KeyError, IndexError, TypeError):
            if isinstance(obj, collections.abc.Sized):
                return len(obj)
            raise

    if key == "first" and not isinstance(obj, str):
        try:
            return (
                await obj.__getitem_async__(key)
                if hasattr(obj, "__getitem_async__")
                else getitem(obj, "first")
            )
        except (KeyError, IndexError, TypeError):
            if isinstance(obj, collections.abc.Mapping) and obj:
                return list(itertools.islice(obj.items(), 1))[0]
            if isinstance(obj, collections.abc.Sequence):
                return obj[0]
            raise

    if key == "last" and not isinstance(obj, str):
        try:
            return (
                await obj.__getitem_async__(key)
                if hasattr(obj, "__getitem_async__")
                else getitem(obj, "last")
            )
        except (KeyError, IndexError, TypeError):
            if isinstance(obj, collections.abc.Sequence):
                return obj[-1]
            raise

    if hasattr(obj, "__getitem_async__"):
        return await obj.__getitem_async__(key)

    return getitem(obj, key)


def get_item(
    obj: Sequence[Any],
    *items: Any,
    default: Optional[object] = _undefined,
) -> Any:
    """Chained item getter."""
    try:
        itm: Any = reduce(_getitem, items, obj)
    except (KeyError, IndexError, TypeError):
        itm = default
    return itm


def is_undefined(obj: object) -> bool:
    """Return `True` if `obj` is undefined. `False` otherwise."""
    return isinstance(obj, Undefined)


class ReadOnlyChainMap(Mapping[str, object]):
    """Combine multiple mappings for sequential lookup.

    Based on the "A greatly simplified read-only version of Chainmap" recipe linked
    to from https://docs.python.org/3/library/collections.html under the section
    chainmap-examples-and-recipes. The link seems to be broken.
    """

    def __init__(self, *maps: Mapping[str, object]):
        self._maps = deque(maps)

    def __getitem__(self, key: str) -> object:
        for mapping in self._maps:
            try:
                return mapping[key]
            except KeyError:
                pass
        raise KeyError(key)

    def __iter__(self) -> Iterator[str]:
        return itertools.chain(*self._maps)

    def __len__(self) -> int:
        return sum(len(map) for map in self._maps)

    def size(self) -> int:
        """Return the number of maps in the chain."""
        return len(self._maps)

    def get(self, key: str, default: object = None) -> object:
        try:
            return self[key]
        except KeyError:
            return default

    def push(self, namespace: Mapping[Any, Any]) -> None:
        """Add a mapping to the front of the chain map."""
        self._maps.appendleft(namespace)

    def pop(self) -> Mapping[Any, Any]:
        """Remove a mapping from the front of the chain map."""
        return self._maps.popleft()


class BuiltIn(Mapping[str, object]):
    """Mapping-like object for resolving built-in, dynamic objects."""

    def __contains__(self, item: object) -> bool:
        if item in ("now", "today"):
            return True
        return False

    def __getitem__(self, key: str) -> object:
        if key == "now":
            return datetime.datetime.now()
        if key == "today":
            return datetime.date.today()
        raise KeyError(str(key))

    def __len__(self) -> int:
        return 2

    def __iter__(self) -> Iterator[str]:
        return iter(["now", "today"])


builtin = BuiltIn()


class Undefined(Mapping[Any, object]):
    """The default undefined type. Always evaluates to an empty string. Can be iterated
    over and indexed without error.
    """

    __slots__ = ("name", "obj", "hint")

    def __init__(self, name: str, obj: object = _undefined, hint: Optional[str] = None):
        self.name = name
        self.obj = obj
        self.hint = hint

    def __contains__(self, item: object) -> bool:
        return False

    def __eq__(self, other: object) -> bool:
        return isinstance(other, Undefined) or other is None

    def __getitem__(self, key: str) -> object:
        return self

    def __len__(self) -> int:
        return 0

    def __iter__(self) -> Iterator[Any]:
        return iter([])

    def __str__(self) -> str:
        return ""

    def __repr__(self) -> str:  # pragma: no cover
        return f"Undefined({self.name})"

    def __int__(self) -> int:
        return 0

    def __hash__(self) -> int:
        return hash(self.name)

    def __reversed__(self) -> Iterable[Any]:
        return []


class DebugUndefined(Undefined):
    """An undefined that returns debug information when rendered."""

    __slots__ = ()

    def __str__(self) -> str:
        if self.hint:
            return f"undefined: {self.hint}"
        if self.obj is not _undefined:
            return f"{type(self.obj).__name__} has no attribute '{self.name}'"
        return f"'{self.name}' is undefined"

    def __repr__(self) -> str:  # pragma: no cover
        return f"Undefined({self.name})"


class StrictUndefined(Undefined):
    """An undefined that raises an exception for everything other than ``repr``."""

    __slots__ = ("msg",)

    # Properties that don't raise an UndefinedError.
    allowed_properties = frozenset(
        [
            "__repr__",
            "name",
            "hint",
            "obj",
            "msg",
        ]
    )

    def __init__(self, name: str, obj: object = _undefined, hint: Optional[str] = None):
        super().__init__(name, obj=obj, hint=hint)
        self.msg = self.hint if self.hint else f"'{self.name}' is undefined"

    def __getattribute__(self, name: str) -> object:
        if name in object.__getattribute__(self, "allowed_properties"):
            return object.__getattribute__(self, name)
        raise UndefinedError(object.__getattribute__(self, "msg"))

    def __contains__(self, item: object) -> bool:
        raise UndefinedError(self.msg)

    def __eq__(self, other: object) -> bool:
        raise UndefinedError(self.msg)

    def __getitem__(self, key: str) -> object:
        raise UndefinedError(self.msg)

    def __len__(self) -> int:
        raise UndefinedError(self.msg)

    def __iter__(self) -> Iterator[Any]:
        raise UndefinedError(self.msg)

    def __str__(self) -> str:
        raise UndefinedError(self.msg)

    def __repr__(self) -> str:
        return f"StrictUndefined({self.name})"

    def __bool__(self) -> bool:
        raise UndefinedError(self.msg)

    def __int__(self) -> int:
        raise UndefinedError(self.msg)

    def __hash__(self) -> int:
        raise UndefinedError(self.msg)

    def __reversed__(self) -> Iterable[Any]:
        raise UndefinedError(self.msg)


class StrictDefaultUndefined(StrictUndefined):
    """An undefined that plays nicely with the `default` filter."""

    # Force the `default` filter to return its default value
    # without inspecting this class type.
    force_liquid_default = True

    # Properties that don't raise an UndefinedError.
    allowed_properties = frozenset(
        [
            "__repr__",
            "force_liquid_default",
            "name",
            "hint",
            "obj",
            "msg",
        ]
    )


# pylint: disable=too-many-instance-attributes redefined-builtin too-many-public-methods
class Context:
    """Liquid template context."""

    __slots__ = (
        "env",
        "locals",
        "globals",
        "counters",
        "scope",
        "loops",
        "tag_namespace",
        "disabled_tags",
        "autoescape",
        "_copy_depth",
        "parent_context",
        "loop_iteration_carry",
        "local_namespace_size_carry",
    )

    getitem = _getitem
    getitem_async = _getitem_async

    # pylint: disable=too-many-arguments
    def __init__(
        self,
        env: Environment,
        globals: Optional[Namespace] = None,
        disabled_tags: Optional[List[str]] = None,
        copy_depth: int = 0,
        parent_context: Optional[Context] = None,
        loop_iteration_carry: int = 1,
        local_namespace_size_carry: int = 0,
    ):
        self.env = env

        # A namespace for template local variables. Those that are bound with
        # `assign` or `capture`.
        self.locals: Dict[str, Any] = {}

        # A read-only namespace containing globally available variables. Usually
        # passed down from the environment.
        self.globals = globals or {}

        # A namespace for `increment` and `decrement` counters.
        self.counters: Dict[str, int] = {}

        # Namespaces are searched in this order. When a context is extended, the
        # temporary namespace is pushed to the front of this chain.
        self.scope = ReadOnlyChainMap(self.locals, self.globals, builtin, self.counters)

        # A namespace supporting stateful tags. Such as `cycle`, `increment`,
        # `decrement` and `ifchanged`.
        self.tag_namespace: Dict[str, Any] = {
            "cycles": {},
            "ifchanged": "",
            "stopindex": {},
        }

        # As stack of forloop objects. Used for populating forloop.parentloop.
        self.loops: List[ForLoop] = []

        # A list of tags names that are disallowed in this context. For example,
        # partial templates rendered using the "render" tag are not allowed to
        # use "include" tags.
        self.disabled_tags = disabled_tags or []

        # Indicates if HTML auto-escaping is enabled.
        self.autoescape = self.env.autoescape

        # Count of the number of times `copy` was called in creating this context. This
        # helps us guard against recursive use of the "render" tag, or at least fail
        # gracefully.
        self._copy_depth = copy_depth

        # A reference to a parent render context, if this one has been copied.
        self.parent_context = parent_context

        # A loop iteration count carried over from a parent context, if this one has
        # been copied. This helps us adhere to loop iteration limits in templates
        # rendered with the `render` tag.
        self.loop_iteration_carry = loop_iteration_carry

        # The cumulative number of names in parent render context local namespaces. We
        # use this to adhere to local namespace limits when copying a render context.
        self.local_namespace_size_carry = local_namespace_size_carry

    def assign(self, key: str, val: Any) -> None:
        """Add `val` to the context with key `key`."""
        self.locals[key] = val
        if (
            self.env.local_namespace_limit
            and self.get_size_of_locals() > self.env.local_namespace_limit
        ):
            raise LocalNamespaceLimitError("local namespace limit reached")

    def get_size_of_locals(self) -> int:
        """Return the "size" or a "score" for the current local namespace.

        This is used by the optional local namespace resource limit. Override
        `get_size_of_locals` to customize how the limit is calculated. Be sure
        to consider `self.local_namespace_size_carry` when writing a custom
        implementation of `get_size_of_locals`.

        The default implementation uses `sys.getsizeof()` on each of the local
        namespace's values. It is not a reliable measure of size in bytes.
        """
        if not self.env.local_namespace_limit:
            return 0
        return (
            sum(sys.getsizeof(obj, default=1) for obj in self.locals.values())
            + self.local_namespace_size_carry
        )

    def get(self, path: ContextPath, default: object = _undefined) -> object:
        """Return the value at path `path` if it is in scope, else default."""
        if isinstance(path, str):
            return self._resolve(path, default)

        name, items = path[0], path[1:]
        assert isinstance(name, str)
        obj = self._resolve(name, default)

        if items:
            try:
                return reduce(Context.getitem, items, obj)
            except (KeyError, IndexError, TypeError) as err:
                if isinstance(err, KeyError):
                    hint = (
                        f"key error: {err}, {name}{''.join([f'[{i}]' for i in items])}"
                    )
                else:
                    hint = f"{err}: {name}{''.join([f'[{i}]' for i in items])}"

                if default == _undefined:
                    return self.env.undefined(
                        name=name,
                        hint=hint,
                    )
                return default

        return obj

    async def get_async(
        self, path: ContextPath, default: object = _undefined
    ) -> object:
        """Return the value at path `path` if it is in scope, else default."""
        if isinstance(path, str):
            return self._resolve(path, default)

        name, items = path[0], path[1:]
        assert isinstance(name, str)
        obj = self._resolve(name, default)

        if items:
            _gi = Context.getitem_async
            try:
                for item in items:
                    obj = await _gi(obj, item)
            except (KeyError, IndexError, TypeError) as err:
                if default == _undefined:
                    return self.env.undefined(
                        name=name,
                        hint=f"{err}: {name}{''.join([f'[{i}]' for i in items])}",
                    )

        return obj

    def resolve(self, name: str, default: object = _undefined) -> Any:
        """Return the object/value at `name` in the current scope.

        This is like `get`, but does a single, top-level lookup rather than a
        chained lookup from a sequence of keys.
        """
        return self._resolve(name, default)

    def _resolve(self, name: str, default: object = _undefined) -> Any:
        try:
            return self.scope[name]
        except KeyError:
            if default == _undefined:
                return self.env.undefined(name)
            return default

    def filter(self, name: str) -> Callable[..., object]:
        """Return the filter function with given name."""
        try:
            filter_func = self.env.filters[name]
        except KeyError as err:
            raise NoSuchFilterFunc(f"unknown filter '{name}'") from err

        kwargs: Dict[str, Any] = {}

        if getattr(filter_func, "with_context", False):
            kwargs["context"] = self

        if getattr(filter_func, "with_environment", False):
            kwargs["environment"] = self.env

        if kwargs:
            return partial(filter_func, **kwargs)

        return filter_func

    def get_template(self, name: str) -> BoundTemplate:
        """Load a template from the environment."""
        return self.env.get_template(name)

    async def get_template_async(self, name: str) -> BoundTemplate:
        """Load a template from the environment asynchronously."""
        return await self.env.get_template_async(name)

    def get_template_with_context(self, name: str, **kwargs: str) -> BoundTemplate:
        """Load a template from the environment, optionally referencing the current
        render context."""
        return self.env.get_template_with_context(self, name, **kwargs)

    async def get_template_with_context_async(
        self,
        name: str,
        **kwargs: str,
    ) -> BoundTemplate:
        """Load a template from the environment asynchronously, optionally referencing
        the current render context."""
        return await self.env.get_template_with_context_async(self, name, **kwargs)

    def increment(self, name: str) -> int:
        """Increment the named counter and return its value."""
        val: int = self.counters.get(name, 0)
        self.counters[name] = val + 1
        return val

    def decrement(self, name: str) -> int:
        """Decrement the named counter and return its value."""
        val: int = self.counters.get(name, 0) - 1
        self.counters[name] = val
        return val

    def cycle(self, group_name: str, args: Sequence[Any]) -> Iterator[Any]:
        """Return the next item in the given cycle. Initialise the cycle first if this
        is the first time we're seeing this combination of group name and arguments."""
        key = (group_name, tuple(args))
        if key not in self.tag_namespace["cycles"]:
            self.tag_namespace["cycles"][key] = cycle(args)
        it: Iterator[Any] = self.tag_namespace["cycles"][key]
        return it

    def ifchanged(self, val: str) -> bool:
        """Return True if the `ifchanged` value has changed."""
        if val != self.tag_namespace["ifchanged"]:
            self.tag_namespace["ifchanged"] = val
            return True

        return False

    def stopindex(self, key: str, index: Optional[int] = None) -> int:
        """Set or return the stop index of a for loop."""
        if index is not None:
            self.tag_namespace["stopindex"][key] = index
            return index

        idx: int = self.tag_namespace["stopindex"].get(key, 0)
        return idx

    @contextmanager
    def extend(self, namespace: Namespace) -> Iterator[Context]:
        """Extend this context with the given read-only namespace."""
        if self.scope.size() > self.env.context_depth_limit:
            raise ContextDepthError(
                "maximum context depth reached, possible recursive include"
            )

        self.scope.push(namespace)

        try:
            yield self
        finally:
            self.scope.pop()

    @contextmanager
    def loop(self, namespace: Namespace, forloop: ForLoop) -> Iterator[Context]:
        """Just like ``Context.extend``, but keeps track of ForLoop objects too."""
        self.raise_for_loop_limit(forloop.length)
        self.loops.append(forloop)
        with self.extend(namespace) as context:
            try:
                yield context
            finally:
                self.loops.pop()

    def parentloop(self) -> Union[Undefined, object]:
        """Return the last ForLoop object from the loop stack."""
        try:
            return self.loops[-1]
        except IndexError:
            return self.env.undefined("parentloop")

    def raise_for_loop_limit(self, length: int = 1) -> None:
        """Raise a ``LoopIterationLimitError`` if the product of the loop stack is
        greater than the configured loop iteration limit.
        """
        if (
            self.env.loop_iteration_limit
            and reduce(
                mul,
                (loop.length for loop in self.loops),
                length * self.loop_iteration_carry,
            )
            > self.env.loop_iteration_limit
        ):
            raise LoopIterationLimitError("loop iteration limit reached")

    def copy(
        self,
        namespace: Namespace,
        disabled_tags: Optional[List[str]] = None,
        carry_loop_iterations: bool = False,
    ) -> Context:
        """Return a copy of this context without any local variables or other state
        for stateful tags."""
        if self._copy_depth > self.env.context_depth_limit:
            raise ContextDepthError(
                "maximum context depth reached, possible recursive render"
            )

        loop_iteration_carry = (
            reduce(
                mul,
                (loop.length for loop in self.loops),
                self.loop_iteration_carry,
            )
            if carry_loop_iterations
            else 1
        )

        return self.__class__(
            self.env,
            globals=ReadOnlyChainMap(namespace, self.globals),
            disabled_tags=disabled_tags,
            copy_depth=self._copy_depth + 1,
            parent_context=self,
            loop_iteration_carry=loop_iteration_carry,
            local_namespace_size_carry=self.get_size_of_locals(),
        )

    def error(self, exc: Error) -> None:
        """Ignore, raise or convert the given exception to a warning."""
        if self.env.mode == Mode.STRICT:
            raise exc
        if self.env.mode == Mode.WARN:
            warnings.warn(str(exc), category=lookup_warning(exc.__class__))

    def get_buffer(self, buf: Optional[TextIO] = None) -> StringIO:
        """Return a new StringIO object, possibly limited according to the configured
        output stream limit."""
        if self.env.output_stream_limit is None:
            return StringIO()

        carry = buf.size if isinstance(buf, LimitedStringIO) else 0
        return LimitedStringIO(limit=self.env.output_stream_limit - carry)


class VariableCaptureContext(Context):
    """A render context that captures template variable names."""

    __slots__ = (
        "local_references",
        "all_references",
        "undefined_references",
        "root_context",
    )

    # Used for formatting context path strings.
    re_ident = re.compile(r"^[\w_][\w_\-]*$")

    # pylint: disable=too-many-arguments
    def __init__(
        self,
        env: Environment,
        globals: Optional[Namespace] = None,
        disabled_tags: Optional[List[str]] = None,
        copy_depth: int = 0,
        parent_context: Optional[VariableCaptureContext] = None,
        loop_iteration_carry: int = 1,
        local_namespace_size_carry: int = 0,
    ):
        super().__init__(
            env,
            globals=globals,
            disabled_tags=disabled_tags,
            copy_depth=copy_depth,
            parent_context=parent_context,
            loop_iteration_carry=loop_iteration_carry,
            local_namespace_size_carry=local_namespace_size_carry,
        )
        self.local_references: List[str] = []
        self.all_references: List[str] = []
        self.undefined_references: List[str] = []

        root_context: VariableCaptureContext = self
        while root_context.parent_context and isinstance(
            root_context.parent_context, VariableCaptureContext
        ):
            root_context = root_context.parent_context
        self.root_context = root_context

    def assign(self, key: str, val: Any) -> None:
        self.local_references.append(key)
        return super().assign(key, val)

    def get(self, path: ContextPath, default: object = _undefined) -> object:
        result = super().get(path, default)
        self._count_reference(path, result)
        return result

    async def get_async(
        self, path: ContextPath, default: object = _undefined
    ) -> object:
        result = await super().get_async(path, default)
        self._count_reference(path, result)
        return result

    def resolve(self, name: str, default: object = _undefined) -> Any:
        result = super().resolve(name, default)
        self._count_reference(name, result)
        return result

    def increment(self, name: str) -> int:
        self.local_references.append(name)
        return super().increment(name)

    def decrement(self, name: str) -> int:
        self.local_references.append(name)
        return super().decrement(name)

    def _count_reference(self, path: ContextPath, result: object) -> None:
        if isinstance(path, str):
            ref = path
        else:
            _path = []
            for elem in path:
                if isinstance(elem, int):
                    _path.append(f"[{elem}]")
                else:
                    if self.re_ident.match(elem):
                        if _path:
                            _path.append(f".{elem}")
                        else:
                            _path.append(f"{elem}")
                    else:
                        _path.append(f'["{elem}"]')
            ref = "".join(_path)

        if is_undefined(result):
            self.root_context.undefined_references.append(ref)
        self.root_context.all_references.append(ref)



"""Malformed template test cases."""
# pylint: disable=missing-class-docstring missing-function-docstring

from unittest import TestCase

from typing import Iterable
from typing import Dict
from typing import List
from typing import Union
from typing import NamedTuple
from typing import Type
from typing import Tuple

from liquid import Template
from liquid.environment import Environment
from liquid.mode import Mode

from liquid.exceptions import LiquidSyntaxError
from liquid.exceptions import FilterArgumentError
from liquid.exceptions import LiquidTypeError
from liquid.exceptions import lookup_warning
from liquid.exceptions import Error
from liquid.exceptions import TemplateNotFound
from liquid.exceptions import DisabledTagError
from liquid.exceptions import ContextDepthError

from liquid.extra import add_inline_expression_tags
from liquid.extra import add_extended_inline_expression_tags
from liquid.extra import IfNotTag

from liquid.loaders import DictLoader


class Case(NamedTuple):
    description: str
    template: str
    expect_exception: Union[Type[Error], Tuple[Type[Error], ...]]
    expect_msg: Union[str, List[str]]
    expect_render: str = ""

    @property
    def exceptions(self):
        if isinstance(self.expect_exception, tuple):
            return self.expect_exception
        return (self.expect_exception,)

    @property
    def warnings(self):
        return tuple(lookup_warning(e) for e in self.exceptions)


class MalformedTemplateTestCase(TestCase):
    """Malformed template test case."""

    def setUp(self):
        self.global_context = {
            "product": {
                "some-tags": ["hello", "there"],
                "tags": ["hello", "there"],
            },
            "tag": "goodbye",
        }

    def _test(self, test_cases: Iterable[Case], mode: Mode = Mode.STRICT):
        """Helper method for running lists of `Case`s in each render mode."""
        self._test_with_env(Environment(tolerance=mode), test_cases)

        # Test with non-standard conditional and boolean expressions
        env = Environment(tolerance=mode)
        env.add_tag(IfNotTag)
        add_inline_expression_tags(env)
        # Skip test cases that are not considered malformed with non-standard
        # expressions.
        self._test_with_env(
            env, [case for case in test_cases if "||" not in case.template]
        )

        # Same again for conditional expressions that support `not` and parens.
        add_extended_inline_expression_tags(env)
        self._test_with_env(
            env, [case for case in test_cases if "||" not in case.template]
        )

    def _test_with_env(self, env: Environment, test_cases: Iterable[Case]):
        for case in test_cases:
            with self.subTest(msg=case.description, mode=env.mode):
                if env.mode == Mode.STRICT:
                    with self.assertRaises(case.exceptions) as raised:
                        template = env.from_string(
                            case.template, globals=self.global_context
                        )
                        template.render()

                    if isinstance(case.expect_msg, list):
                        self.assertIn(str(raised.exception), case.expect_msg)
                    else:
                        self.assertEqual(str(raised.exception), case.expect_msg)

                elif env.mode == Mode.WARN:
                    with self.assertWarns(case.warnings):
                        template = env.from_string(
                            case.template, globals=self.global_context
                        )
                        template.render()
                elif env.mode == Mode.LAX:
                    template = env.from_string(
                        case.template, globals=self.global_context
                    )
                    result = template.render()
                    self.assertEqual(result, case.expect_render)

    def _test_partial(self, test_cases: Iterable[Case], templates: Dict[str, str]):
        """Helper method for testing lists of 'include' or 'render' cases."""
        env = Environment(loader=DictLoader(templates))
        for case in test_cases:
            with self.subTest(msg=case.description):
                with self.assertRaises(case.exceptions) as raised:
                    template = env.from_string(
                        case.template, globals=self.global_context
                    )
                    template.render()

                if isinstance(case.expect_msg, list):
                    self.assertIn(str(raised.exception), case.expect_msg)
                else:
                    self.assertEqual(str(raised.exception), case.expect_msg)

        env = Environment(loader=DictLoader(templates), tolerance=Mode.WARN)
        for case in test_cases:
            with self.subTest(msg=case.description):
                with self.assertWarns(case.warnings):
                    template = env.from_string(
                        case.template, globals=self.global_context
                    )
                    template.render()

        env = Environment(loader=DictLoader(templates), tolerance=Mode.LAX)
        for case in test_cases:
            with self.subTest(msg=case.description):
                template = env.from_string(case.template, globals=self.global_context)
                result = template.render()
                self.assertEqual(result, case.expect_render)

    def test_liquid_syntax(self):
        """Test that we fail early and loud when parsing a malformed template."""

        test_cases = [
            Case(
                description="no expression",
                template="{% if %}foo{% endif %}",
                expect_exception=LiquidSyntaxError,
                expect_msg="expected 'expression', found 'literal', on line 1",
            ),
            Case(
                description="end tag mismatch",
                template="{% if true %}foo{% endunless %}",
                expect_exception=LiquidSyntaxError,
                expect_msg="unexpected tag 'endunless', on line 1",
            ),
            Case(
                description="unexpected outer tag name",
                template="{% foo true %}foo{% endfoo %}",
                expect_exception=LiquidSyntaxError,
                expect_msg="unexpected tag 'foo', on line 1",
            ),
            Case(
                description="unexpected tag name inside block with line numbers",
                template=("Hello\n\n{% if true %}\n{% foo %}{% endfoo %}\n{% endif %}"),
                expect_exception=LiquidSyntaxError,
                expect_msg="unexpected tag 'foo', on line 4",
                expect_render="Hello\n\n",
            ),
            Case(
                description="missing tag name",
                template="{% %}foo{% end %}",
                expect_exception=LiquidSyntaxError,
                expect_msg="unexpected tag '', on line 1",
            ),
            Case(
                description="missing end tag at EOF",
                template="{% if true %}foo{% assign bar = 'baz' %}",
                expect_exception=LiquidSyntaxError,
                expect_msg=(
                    "expected tag with value 'endif', found eof with value '', "
                    "on line 1"
                ),
            ),
            Case(
                description="orphaned break",
                template="{% break %}",
                expect_exception=LiquidSyntaxError,
                expect_msg="unexpected 'break', on line 1",
            ),
            Case(
                description="orphaned continue",
                template="{% continue %}",
                expect_exception=LiquidSyntaxError,
                expect_msg="unexpected 'continue', on line 1",
            ),
            Case(
                description="orphaned else",
                template="{% else %}",
                expect_exception=LiquidSyntaxError,
                expect_msg="unexpected tag 'else', on line 1",
            ),
            Case(
                description="orphaned when",
                template="{% when %}",
                expect_exception=LiquidSyntaxError,
                expect_msg="unexpected tag 'when', on line 1",
            ),
            Case(
                description="missing 'in' in forloop",
                template="{% for x (0..3) %}{{ x }}{% endfor %}",
                expect_exception=LiquidSyntaxError,
                expect_msg="expected 'in', found '(', on line 1",
            ),
            Case(
                description="missing range or identifier in forloop",
                template="{% for x in %}{{ x }}foo{% endfor %}",
                expect_exception=LiquidSyntaxError,
                expect_msg="invalid loop expression, on line 1",
            ),
            Case(
                description="float with trailing dot in range literal",
                template="{% for x in (2...4) %}{{ x }}{% endfor %}",
                expect_exception=LiquidSyntaxError,
                expect_msg=[
                    "invalid range expression, expected an integer, "
                    "found a dot, on line 1",
                    "unexpected '.' in range expression, on line 1",
                ],
            ),
            Case(
                description="chained identifier for loop variable",
                template="{% for x.y in (2...4) %}{{ x }}{% endfor %}",
                expect_exception=LiquidSyntaxError,
                expect_msg="expected 'in', found '.', on line 1",
            ),
            Case(
                description="missing equal in assignment tag",
                template="{% assign x 5 %}",
                expect_exception=LiquidSyntaxError,
                expect_msg='invalid assignment expression "x 5", on line 1',
            ),
            Case(
                description="invalid subscript identifier",
                template="{{ foo[1.2] }}",
                expect_exception=LiquidSyntaxError,
                expect_msg=[
                    "invalid identifier, found float, on line 1",
                    "expected 'identifier', found 'float', on line 1",
                ],
            ),
            Case(
                description="nil subscript identifier",
                template="{{ foo[nil] }}",
                expect_exception=LiquidSyntaxError,
                expect_msg=[
                    "invalid identifier, found nil, on line 1",
                    "expected 'identifier', found 'nil', on line 1",
                ],
            ),
            Case(
                description="minus string",
                template="{{ -'foo' }}",
                expect_exception=(LiquidTypeError, LiquidSyntaxError),
                expect_msg=[
                    "unknown operator -'foo', on line 1",
                    "unexpected '-', on line 1",
                ],
            ),
            Case(
                description="unknown prefix operator",
                template="{{ +5 }}",
                expect_exception=LiquidSyntaxError,
                expect_msg="unexpected '+', on line 1",
            ),
            Case(
                description="float literal without a leading zero",
                template="{{ .1 }}",
                expect_exception=LiquidSyntaxError,
                expect_msg=[
                    "unknown prefix operator '.', on line 1",
                    "unexpected '.', on line 1",
                ],
            ),
            Case(
                description="unknown infix operator",
                template="{% if 1 =! 2 %}ok{% endif %}",
                expect_exception=LiquidSyntaxError,
                expect_msg="unknown operator '=!', on line 1",
            ),
            Case(
                description="bad 'unless' expression",
                template="{% unless 1 ~ 2 %}ok{% endunless %}",
                expect_exception=LiquidSyntaxError,
                expect_msg="unexpected '~', on line 1",
            ),
            Case(
                description="bad conditional expression in unless block",
                template="{% unless true %}ok{% elsif £$! %}{% endunless %}",
                expect_exception=LiquidSyntaxError,
                expect_msg="unexpected '£', on line 1",
            ),
            Case(
                description="unknown infix operator",
                template="{% if 1 ~ 2 %}ok{% endif %}",
                expect_exception=LiquidSyntaxError,
                expect_msg="unexpected '~', on line 1",
            ),
            Case(
                description="bad alternative condition expression",
                template="{% if false %}ok{% elsif 1~=2 %}not ok{% endif %}",
                expect_exception=LiquidSyntaxError,
                expect_msg="unexpected '~', on line 1",
            ),
            Case(
                description="junk in `liquid` tag",
                template="\n".join(
                    [
                        r"{{ 'hello' }}",
                        r"{% liquid",
                        r"echo 'foo'",
                        r"aiu34bseu",
                        r"%}",
                    ]
                ),
                expect_exception=LiquidSyntaxError,
                expect_render="hello\n",
                expect_msg="unexpected tag 'aiu34bseu', on line 4",
            ),
            Case(
                description="bad token in loop expression",
                template="{% for i$ in (1..3) %}{% endfor %}",
                expect_exception=LiquidSyntaxError,
                expect_msg="unexpected '$', on line 1",
            ),
            Case(
                description="invalid loop argument",
                template=(
                    "{% for product in collections[0]['tags'] limit:| %}{% endfor %}"
                ),
                expect_exception=LiquidSyntaxError,
                expect_msg="unexpected '|', on line 1",
            ),
            Case(
                description="double pipe characters in filtered expression",
                template=r'{{ "failure" || upcase }}',
                expect_exception=LiquidSyntaxError,
                expect_msg="unexpected pipe or missing filter name, on line 1",
            ),
        ]

        self._test(test_cases, mode=Mode.STRICT)
        self._test(test_cases, mode=Mode.WARN)
        self._test(test_cases, mode=Mode.LAX)

    def test_liquid_syntax_from_template_api(self):
        """Test that syntax errors are raised by default when using the `Template`
        API."""
        with self.assertRaises(LiquidSyntaxError):
            # Missing colon before filter argument
            Template(r"{{ a | sort foo }}")

    def test_unrecoverable_syntax_errors(self):
        """Test that we fail early and loud when parsing a malformed template."""

        test_cases = [
            Case(
                description="single bracket close",
                template=r"text {{method} oh nos!",
                expect_exception=LiquidSyntaxError,
                expect_msg="expected '}}', found 'eof', on line 1",
            ),
            Case(
                description="single bracket close from tag sequence",
                template=r"text {%method} oh nos!",
                expect_exception=LiquidSyntaxError,
                expect_msg="expected '%}', found 'eof', on line 1",
            ),
        ]

        self._test(test_cases, mode=Mode.STRICT)

    def test_bad_include(self):
        """Test that we gracefully handle include errors."""

        test_cases = [
            Case(
                description="chained alias identifier",
                template=(
                    "{% include 'product.alt' with product['some-tags'] as foo.bar %}"
                ),
                expect_exception=LiquidSyntaxError,
                expect_msg="invalid identifier 'foo.bar', on line 1",
            ),
            Case(
                description="chained keyword identifier",
                template="{% include 'product.alt', foo.bar: 'hello' %}",
                expect_exception=LiquidSyntaxError,
                expect_msg="invalid identifier 'foo.bar', on line 1",
            ),
            Case(
                description="template not found",
                template="{% include 'no/such/thing' %}",
                expect_exception=TemplateNotFound,
                expect_msg="template not found no/such/thing, on line 1",
            ),
            Case(
                description="template not found carry on",
                template="hello {% include 'no/such/thing' %}there",
                expect_exception=TemplateNotFound,
                expect_msg="template not found no/such/thing, on line 1",
                expect_render="hello there",
            ),
            Case(
                description="template name not a string or identifier",
                template="{% include 5 %}",
                expect_exception=LiquidSyntaxError,
                expect_msg="expected identifier or string, found integer, on line 1",
            ),
            Case(
                description="missing keyword argument separator",
                template="{% include 'product.alt', foo: 5 bar: 6 %}",
                expect_exception=LiquidSyntaxError,
                expect_msg=(
                    "expected a comma separated list of arguments, "
                    "found identifier, on line 1"
                ),
            ),
            Case(
                description="bound expression is not an identifier",
                template="{% include 'product.alt' with 'foo' %}",
                expect_exception=LiquidSyntaxError,
                expect_msg="expected 'identifier', found 'string', on line 1",
            ),
            Case(
                description="recursive include",
                template="{% include 'recursive.liquid' %}",
                expect_exception=ContextDepthError,
                expect_msg=(
                    "maximum context depth reached, "
                    "possible recursive include, on line 1"
                ),
            ),
        ]

        templates = {
            "product.alt": "{{ product }}",
            "recursive.liquid": "{% include 'inner.liquid' %}",
            "inner.liquid": "{% include 'recursive.liquid' %}",
        }

        self._test_partial(test_cases, templates)

    def test_bad_render(self):
        """Test that we gracefully handle render errors."""

        test_cases = [
            Case(
                description="break from render",
                template="".join(
                    [
                        r"{% assign greeting = 'good morning' %}",
                        r"{% for i in (1..3) %}",
                        r"{{ i }}",
                        r"{% render 'break' %}",
                        r"{{ i }}",
                        r"{% endfor %}",
                    ]
                ),
                expect_exception=LiquidSyntaxError,
                expect_msg="unexpected 'break', on line 1",
                expect_render="112233",
            ),
            Case(
                description="template name from identifier",
                template="{% render foo.bar %}",
                expect_exception=LiquidSyntaxError,
                expect_msg="expected 'string', found 'identifier', on line 1",
            ),
            Case(
                description="include will not render inside render",
                template="{% render 'include' %}",
                expect_exception=DisabledTagError,
                expect_msg="include usage is not allowed in this context, on line 1",
            ),
            Case(
                description="recursive render",
                template="{% render 'recursive.liquid' %}",
                expect_exception=ContextDepthError,
                expect_msg=(
                    "maximum context depth reached, "
                    "possible recursive render, on line 1"
                ),
            ),
            Case(
                description="arguments are not comma separated",
                template="{% render 'index', foo: 'foo' bar: 'bar' %}",
                expect_exception=LiquidSyntaxError,
                expect_msg=(
                    "expected a comma separated list of arguments, "
                    "found identifier, on line 1"
                ),
            ),
        ]

        templates = {
            "index": "hello",
            "break": "{% break %}",
            "include": "{% include 'foo' %}",
            "recursive.liquid": "{% render 'inner.liquid' %}",
            "inner.liquid": "{% render 'recursive.liquid' %}",
        }

        self._test_partial(test_cases, templates)

    def test_resume_block(self):
        """Test that we continue to execute a block after a single statement error."""
        source = (
            r"{% if true %}"
            r"before error "
            r"{% comment  %}"
            r"The following filter expression should throw a FilterArgumentError."
            r"{% endcomment %}"
            r"{{ 'foo' | upcase: bad }}"
            r"after error"
            r"{% endif %}"
        )

        env = Environment()

        # Confirm the exception is raised when in strict mode.
        env.mode = Mode.STRICT

        template = env.from_string(source)
        with self.assertRaises(FilterArgumentError):
            template.render()

        # Expect the template literal at the end of the block to render
        # after the filter argument error.
        env.mode = Mode.LAX
        result = template.render()
        self.assertEqual(result, "before error after error")

        # Same, but warn too.
        env.mode = Mode.WARN
        with self.assertWarns(lookup_warning(FilterArgumentError)):
            result = template.render()

        self.assertEqual(result, "before error after error")

    def test_invalid_identifiers(self):
        """Test that we gracefully handle invalid identifiers."""
        test_cases = [
            Case(
                description="unexpected character",
                template=r"{% assign foo+bar = 'hello there'%}{{ foo+bar }}",
                expect_exception=LiquidSyntaxError,
                expect_msg=(
                    "invalid assignment expression \"foo+bar = 'hello there'\", "
                    "on line 1"
                ),
            ),
            Case(
                description="unexpected chained assignment identifier",
                template=r"{% assign foo.bar = 'hello there' %}{{ foo.bar }}",
                expect_exception=LiquidSyntaxError,
                expect_msg=(
                    "invalid assignment expression \"foo.bar = 'hello there'\", "
                    "on line 1"
                ),
            ),
            Case(
                description="unexpected chained capture identifier",
                template=r"{% capture foo.bar %}{% endcapture %}",
                expect_exception=LiquidSyntaxError,
                expect_msg='invalid capture identifier "foo.bar", on line 1',
            ),
        ]

        self._test(test_cases, mode=Mode.STRICT)
        # self._test(test_cases, mode=Mode.WARN)
        # self._test(test_cases, mode=Mode.LAX)



"""Golden tests cases for testing liquid's `decrement` tag."""

from liquid.golden.case import Case

cases = [
    Case(
        description="named counter",
        template=r"{% decrement foo %}{{ foo }} {% decrement foo %}{{ foo }}",
        expect="-1-1 -2-2",
    ),
    Case(
        description="increment and decrement named counter",
        template=r"{% decrement foo %} {% decrement foo %} {% increment foo %}",
        expect="-1 -2 -2",
    ),
]



"""Golden tests cases for testing liquid's `case` tag."""

from liquid.golden.case import Case

cases = [
    Case(
        description="simple case/when",
        template=(
            r"{% case title %}"
            r"{% when 'foo' %}foo"
            r"{% when 'Hello' %}bar"
            r"{% endcase %}"
        ),
        expect="bar",
        globals={"title": "Hello"},
    ),
    Case(
        description="'when' expression using an identifier",
        template=(
            r"{% case title %}"
            r"{% when other %}foo"
            r"{% when 'goodbye' %}bar"
            r"{% endcase %}"
        ),
        expect="foo",
        globals={"title": "Hello", "other": "Hello"},
    ),
    Case(
        description="tags inside when block",
        template=(
            r"{% case title %}"
            r"{% when other %}"
            r"{% if true %}foo{% endif %}"
            r"{% when 'goodbye' %}bar"
            r"{% endcase %}"
        ),
        expect="foo",
        globals={"title": "Hello", "other": "Hello"},
    ),
    Case(
        description="'when' expression using an out of scope identifier",
        template=(
            r"{% case title %}"
            r"{% when nosuchthing %}foo"
            r"{% when 'Hello' %}bar"
            r"{% endcase %}"
        ),
        expect="bar",
        globals={"title": "Hello"},
    ),
    Case(
        description="name not in scope",
        template=(
            r"{% case nosuchthing %}"
            r"{% when 'foo' %}foo"
            r"{% when 'bar' %}bar"
            r"{% endcase %}"
        ),
        expect="",
    ),
    Case(
        description="no match and no default",
        template=(
            r"{% case title %}"
            r"{% when 'foo' %}foo"
            r"{% when 'bar' %}bar"
            r"{% endcase %}"
        ),
        expect="",
        globals={"title": "Hello"},
    ),
    Case(
        description="with default",
        template=r"{% case title %}{% when 'foo' %}foo{% else %}bar{% endcase %}",
        expect="bar",
        globals={"title": "Hello"},
    ),
    Case(
        description="no whens",
        template=r"{% case title %}{% else %}bar{% endcase %}",
        expect="bar",
        globals={"title": "Hello"},
    ),
    Case(
        description="no whens or default",
        template=r"{% case title %}{% endcase %}",
        expect="",
        globals={"title": "Hello"},
    ),
    Case(
        description="whitespace",
        template=(
            "{% case title %}  \n\t"
            "{% when 'foo' %}foo\n"
            "{% when 'Hello' %}bar"
            "{% endcase %}"
        ),
        expect="bar",
        globals={"title": "Hello"},
    ),
    Case(
        description="comma separated when expression",
        template=(
            r"{% case title %}"
            r"{% when 'foo' %}foo"
            r"{% when 'bar', 'Hello' %}bar"
            r"{% endcase %}"
        ),
        expect="bar",
        globals={"title": "Hello"},
    ),
    Case(
        description="evaluate multiple matching blocks",
        template=(
            r"{% case title %}"
            r"{% when 'Hello' %}foo"
            r"{% when a, 'Hello' %}bar"
            r"{% endcase %}"
        ),
        expect="foobarbar",
        globals={"title": "Hello", "a": "Hello"},
    ),
]



"""Tag and node definition for the built-in "comment" tag."""

import sys

from typing import List
from typing import Optional
from typing import TextIO

from liquid import ast
from liquid.context import Context
from liquid.stream import TokenStream
from liquid.tag import Tag

from liquid.parse import expect
from liquid.parse import eat_block

from liquid.token import Token
from liquid.token import TOKEN_EOF
from liquid.token import TOKEN_TAG


TAG_COMMENT = sys.intern("comment")
TAG_ENDCOMMENT = sys.intern("endcomment")

END_COMMENTBLOCK = frozenset((TAG_ENDCOMMENT,))


class CommentNode(ast.Node):
    """Parse tree node for the built-in "comment" tag."""

    __slots__ = ("tok", "text")

    def __init__(self, tok: Token, text: Optional[str] = None):
        self.tok = tok
        self.text = text

    def __str__(self) -> str:
        if self.text:
            return f"/* {self.text} */"
        return "/* */"

    def render_to_output(
        self,
        context: Context,
        buffer: TextIO,
    ) -> Optional[bool]:
        return False

    def children(self) -> List[ast.ChildNode]:
        return []


class CommentTag(Tag):
    """The built-in "comment" tag.

    This implementation does not include comment text in the resulting
    AST node.
    """

    name = TAG_COMMENT
    end = TAG_ENDCOMMENT

    def parse(self, stream: TokenStream) -> CommentNode:
        expect(stream, TOKEN_TAG, value=TAG_COMMENT)
        stream.next_token()
        node = CommentNode(stream.current)
        eat_block(stream, end=END_COMMENTBLOCK)
        return node


class CommentTextTag(CommentTag):
    """An implementation of the built-in "comment" tag that includes
    comment text in a template's AST.

    Some Liquid markup might be stripped out by the lexer, so comment
    text is not guaranteed to be identical to that in the source document.
    """

    name = TAG_COMMENT
    end = TAG_ENDCOMMENT

    def parse(self, stream: TokenStream) -> CommentNode:
        expect(stream, TOKEN_TAG, value=TAG_COMMENT)
        stream.next_token()
        tok = stream.current

        text = []
        while stream.current.type != TOKEN_EOF:
            if (
                stream.current.type == TOKEN_TAG
                and stream.current.value == TAG_ENDCOMMENT
            ):
                break
            text.append(stream.current.value)
            stream.next_token()

        return CommentNode(tok, text="".join(text))



"""Tokenize liquid filtered expressions.

Filtered expressions are those found in the output statement, the
`assign` tag and the `echo` tag.
"""
import re
from typing import Iterator

from liquid.expressions.common import Token
from liquid.expressions.common import GROUP_QUOTED
from liquid.expressions.common import GROUP_IDENTINDEX
from liquid.expressions.common import GROUP_IDENTQUOTED
from liquid.expressions.common import IDENTIFIER_PATTERN
from liquid.expressions.common import IDENTSTRING_PATTERN
from liquid.expressions.common import IDENTINDEX_PATTERN
from liquid.expressions.common import STRING_PATTERN

from liquid.token import TOKEN_RANGE
from liquid.token import TOKEN_FLOAT
from liquid.token import TOKEN_INTEGER
from liquid.token import TOKEN_TRUE
from liquid.token import TOKEN_FALSE
from liquid.token import TOKEN_NIL
from liquid.token import TOKEN_NULL
from liquid.token import TOKEN_EMPTY
from liquid.token import TOKEN_BLANK
from liquid.token import TOKEN_IDENTIFIER
from liquid.token import TOKEN_STRING
from liquid.token import TOKEN_ILLEGAL
from liquid.token import TOKEN_SKIP
from liquid.token import TOKEN_NEWLINE
from liquid.token import TOKEN_LBRACKET
from liquid.token import TOKEN_RBRACKET
from liquid.token import TOKEN_LPAREN
from liquid.token import TOKEN_RPAREN
from liquid.token import TOKEN_COLON
from liquid.token import TOKEN_PIPE
from liquid.token import TOKEN_COMMA
from liquid.token import TOKEN_IDENTSTRING
from liquid.token import TOKEN_IDENTINDEX
from liquid.token import TOKEN_DOT

from liquid.exceptions import LiquidSyntaxError


token_rules = (
    (TOKEN_IDENTINDEX, IDENTINDEX_PATTERN),
    (TOKEN_IDENTSTRING, IDENTSTRING_PATTERN),
    (TOKEN_STRING, STRING_PATTERN),
    (TOKEN_RANGE, r"\.\."),
    (TOKEN_FLOAT, r"-?\d+\.(?!\.)\d*"),
    (TOKEN_INTEGER, r"-?\d+\b"),
    (TOKEN_DOT, r"\."),
    (TOKEN_IDENTIFIER, IDENTIFIER_PATTERN),
    (TOKEN_LPAREN, r"\("),
    (TOKEN_RPAREN, r"\)"),
    (TOKEN_LBRACKET, r"\["),
    (TOKEN_RBRACKET, r"]"),
    (TOKEN_COMMA, r","),
    (TOKEN_COLON, r":"),
    (TOKEN_PIPE, r"\|"),
    (TOKEN_NEWLINE, r"\n"),
    (TOKEN_SKIP, r"[ \t\r]+"),
    (TOKEN_ILLEGAL, r"."),
)

keywords = frozenset(
    [
        TOKEN_TRUE,
        TOKEN_FALSE,
        TOKEN_NIL,
        TOKEN_NULL,
        TOKEN_EMPTY,
        TOKEN_BLANK,
    ]
)

OUTPUT_RE = re.compile(
    "|".join(f"(?P<{name}>{pattern})" for name, pattern in token_rules),
    re.DOTALL,
)


def tokenize(source: str, linenum: int = 1) -> Iterator[Token]:
    """Yield tokens from an output expression."""
    _keywords = keywords
    for match in OUTPUT_RE.finditer(source):
        kind = match.lastgroup
        assert kind is not None

        value = match.group()
        newlines = value.count("\n")

        if kind == TOKEN_IDENTIFIER and value in _keywords:
            kind = value
        elif kind == TOKEN_IDENTINDEX:
            value = match.group(GROUP_IDENTINDEX)
        elif kind == TOKEN_IDENTSTRING:
            kind = TOKEN_IDENTIFIER
            value = match.group(GROUP_IDENTQUOTED)
        elif kind == TOKEN_STRING:
            value = match.group(GROUP_QUOTED)
        elif kind == TOKEN_NEWLINE:
            linenum += 1
            continue
        elif kind == TOKEN_SKIP:
            continue
        elif kind == TOKEN_ILLEGAL:
            raise LiquidSyntaxError(f"unexpected {value!r}", linenum=linenum)

        linenum += newlines
        yield (linenum, kind, value)



"""Tag and node definition for the built-in "increment" tag."""

import sys

from typing import List
from typing import Optional
from typing import TextIO

from liquid.ast import ChildNode
from liquid.ast import Node

from liquid.context import Context
from liquid.stream import TokenStream
from liquid.tag import Tag
from liquid.parse import expect

from liquid.token import Token
from liquid.token import TOKEN_TAG
from liquid.token import TOKEN_EXPRESSION

from liquid.expressions import TokenStream as ExprTokenStream
from liquid.expressions.common import parse_unchained_identifier
from liquid.expressions.filtered.lex import tokenize

TAG_INCREMENT = sys.intern("increment")


class IncrementNode(Node):
    """Parse tree node for the built-in "increment" tag."""

    __slots__ = ("tok", "identifier")

    def __init__(self, tok: Token, identifier: str):
        self.tok = tok
        self.identifier = identifier

    def __str__(self) -> str:
        return f"{self.identifier} += 1"

    def render_to_output(self, context: Context, buffer: TextIO) -> Optional[bool]:
        buffer.write(str(context.increment(self.identifier)))
        return True

    def children(self) -> List[ChildNode]:
        return [ChildNode(linenum=self.tok.linenum, template_scope=[self.identifier])]


class IncrementTag(Tag):
    """The built-in "increment" tag."""

    name = TAG_INCREMENT
    block = False

    def parse(self, stream: TokenStream) -> IncrementNode:
        expect(stream, TOKEN_TAG, value=TAG_INCREMENT)
        tok = stream.current
        stream.next_token()
        expect(stream, TOKEN_EXPRESSION)
        return IncrementNode(
            tok=tok,
            identifier=str(
                parse_unchained_identifier(
                    ExprTokenStream(tokenize(stream.current.value))
                )
            ),
        )



"""Golden tests cases for testing liquid's built-in `remove` filter."""

from liquid.golden.case import Case

cases = [
    Case(
        description="remove substrings",
        template=(
            r'{{ "I strained to see the train through the rain" | remove: "rain" }}'
        ),
        expect="I sted to see the t through the ",
    ),
    Case(
        description="not a string",
        template=r"{{ 5 | remove: 'there' }}",
        expect="5",
    ),
    Case(
        description="argument not a string",
        template=r'{{ "hello" | remove: 5 }}',
        expect="hello",
    ),
    Case(
        description="missing argument",
        template=r'{{ "hello" | remove }}',
        expect="",
        error=True,
    ),
    Case(
        description="too many arguments",
        template=r'{{ "hello" | remove: "how", "are", "you" }}',
        expect="",
        error=True,
    ),
    Case(
        description="undefined left value",
        template=r'{{ nosuchthing | remove: "rain" }}',
        expect="",
    ),
    Case(
        description="undefined argument",
        template=r'{{ "hello" | remove: nosuchthing }}',
        expect="hello",
    ),
]



"""Golden tests cases for testing liquid's `include` tag."""

from liquid.golden.case import Case
from liquid.golden.case import TEMPLATE_DROP_ATTRS

cases = [
    Case(
        description="string literal name",
        template=r"{% include 'product-hero' %}",
        expect="foo\n- sports\n- garden\n",
        globals={"product": {"title": "foo", "tags": ["sports", "garden"]}},
        partials={
            "product-hero": (
                r"{{ product.title }}"
                "\n"
                r"{% for tag in product.tags %}"
                r"- {{ tag }}"
                "\n"
                r"{% endfor %}"
            ),
        },
    ),
    Case(
        description="name from identifier",
        template=r"{% include snippet %}",
        expect="foo\n- sports\n- garden\n",
        globals={
            "snippet": "product-hero",
            "product": {"title": "foo", "tags": ["sports", "garden"]},
        },
        partials={
            "product-hero": (
                r"{{ product.title }}"
                "\n"
                r"{% for tag in product.tags %}"
                r"- {{ tag }}"
                "\n"
                r"{% endfor %}"
            ),
        },
    ),
    Case(
        description="bound variable",
        template=r"{% include 'product-title' with collection.products[1] %}",
        expect="car",
        globals={
            "collection": {
                "products": [{"title": "bike"}, {"title": "car"}],
            }
        },
        partials={
            "product-title": r"{{ product-title.title }}",
        },
    ),
    Case(
        description="bound variable does not exist",
        template=r"{% include 'product-title' with no.such.thing %}",
        expect="",
        partials={
            "product-title": r"{{ product-title.title }}",
        },
    ),
    Case(
        description="bound array variable",
        template=r"{% include 'prod' for collection.products %}",
        expect="bikecar",
        globals={
            "collection": {
                "products": [{"title": "bike"}, {"title": "car"}],
            }
        },
        partials={"prod": r"{{ prod.title }}"},
    ),
    Case(
        description="bound variable with alias",
        template=(
            r"{% include 'product-alias' with collection.products[1] as product %}"
        ),
        expect="car",
        globals={
            "collection": {
                "products": [{"title": "bike"}, {"title": "car"}],
            }
        },
        partials={"product-alias": r"{{ product.title }}"},
    ),
    Case(
        description="some keyword arguments",
        template=r"{% include 'product-args', foo: 'hello', bar: 'there' %}",
        expect="hello there",
        partials={"product-args": r"{{ foo }} {{ bar }}"},
    ),
    Case(
        description="some keyword arguments without leading comma",
        template=r"{% include 'product-args' foo: 'hello', bar: 'there' %}",
        expect="hello there",
        partials={"product-args": r"{{ foo }} {{ bar }}"},
    ),
    Case(
        description="some keyword arguments with float literals",
        template=r"{% include 'product-args' foo: 1.1, bar: 'there' %}",
        expect="1.1 there",
        partials={"product-args": r"{{ foo }} {{ bar }}"},
    ),
    Case(
        description="some keyword arguments with range literal",
        template=r"{% include 'product-args' foo: (1..3), bar: 'there' %}",
        expect="1#2#3 there",
        partials={"product-args": r"{{ foo | join: '#' }} {{ bar }}"},
    ),
    Case(
        description="template drop",
        template=r"{% include 'some/template-attrs.alt.txt' %}",
        expect="template-attrs.alt some template-attrs alt",
        partials={"some/template-attrs.alt.txt": TEMPLATE_DROP_ATTRS},
        standard=False,
    ),
    Case(
        description="template drop no parent",
        template=r"{% include 'template-attrs.alt.txt' %}",
        expect="template-attrs.alt  template-attrs alt",
        partials={"template-attrs.alt.txt": TEMPLATE_DROP_ATTRS},
        standard=False,
    ),
    Case(
        description="template drop no suffix",
        template=r"{% include 'some/template-attrs.txt' %}",
        expect="template-attrs some template-attrs ",
        partials={"some/template-attrs.txt": TEMPLATE_DROP_ATTRS},
        standard=False,
    ),
    Case(
        description="template drop no suffix or extension",
        template=r"{% include 'some/template-attrs' %}",
        expect="template-attrs some template-attrs ",
        partials={"some/template-attrs": TEMPLATE_DROP_ATTRS},
        standard=False,
    ),
    Case(
        description="use globals from outer scope",
        template=r"{% include 'outer-scope' %}",
        expect="Hello, Holly",
        globals={"customer": {"first_name": "Holly"}},
        partials={"outer-scope": r"Hello, {{ customer.first_name }}"},
    ),
    Case(
        description="assign persists in outer scope",
        template=r"{% include 'assign-outer-scope' %} {{ last_name }}",
        expect="Hello, Holly Smith",
        globals={"customer": {"first_name": "Holly"}},
        partials={
            "assign-outer-scope": (
                r"Hello, {{ customer.first_name }}{% assign last_name = 'Smith' %}"
            ),
        },
    ),
    Case(
        description="counter from outer scope",
        template=(
            r"{% increment foo %} "
            r"{% include 'increment-outer-scope' %} "
            r"{% increment foo %}"
        ),
        expect="0 1 2",
        partials={"increment-outer-scope": r"{% increment foo %}"},
    ),
    Case(
        description="break from include",
        template=r"{% for tag in product.tags %}{% include 'tag-break' %}{% endfor %}",
        expect="SPORTS",
        globals={"product": {"tags": ["sports", "garden"]}},
        partials={"tag-break": r"{{ tag | upcase }}{% break %}"},
    ),
    Case(
        description="break from nested include",
        template=r"{% for tag in product.tags %}{% include 'tag' %}{% endfor %}",
        expect="SPORTS break!",
        globals={"product": {"tags": ["sports", "garden"]}},
        partials={
            "tag": r"{{ tag | upcase }}{% include 'break' %}",
            "break": r" break!{% break %}",
        },
    ),
    Case(
        description="keyword arguments go out of scope",
        template=r"{% include 'product-args', foo: 'hello', bar: 'there' %}{{ foo }}",
        expect="hello there",
        partials={"product-args": r"{{ foo }} {{ bar }}"},
    ),
    Case(
        description="assign to a keyword argument",
        template=r"{% include 'product-args', foo: 'hello' %} {{ foo }}",
        expect="hello hello goodbye",
        partials={"product-args": r"{{ foo }}{% assign foo = 'goodbye' %} {{ foo }}"},
    ),
]



"""Base class and file system implementation for loading template sources.

Modelled after Jinja2 template loaders.
See https://github.com/pallets/jinja/blob/master/src/jinja2/loaders.py
"""
from __future__ import annotations

import asyncio
import os

from abc import ABC

from collections import abc
from functools import partial
from pathlib import Path

from typing import Awaitable
from typing import NamedTuple
from typing import Callable
from typing import Dict
from typing import Iterable
from typing import List
from typing import Mapping
from typing import Optional
from typing import Tuple
from typing import Union
from typing import TYPE_CHECKING

from liquid.template import BoundTemplate
from liquid.exceptions import TemplateNotFound


if TYPE_CHECKING:  # pragma: no cover
    from liquid import Environment
    from liquid import Context

UpToDate = Union[Callable[[], bool], Callable[[], Awaitable[bool]], None]


class TemplateSource(NamedTuple):
    """A Liquid template source as returned by the ``get_source`` method of a `loader`.

    :param source: The liquid template source code.
    :type source: str
    :param filename: The liquid template file name or other string identifying its
        origin.
    :type filename: str
    :param uptodate: Optional callable that will return ``True`` if the template is up
        to date, or ``False`` if it needs to be reloaded. Defaults to ``None``.
    :type uptodate: Union[Callable[[], bool], Callable[[], Awaitable[bool]], None]
    :param matter: Optional mapping containing variables associated with the template.
        Could be "front matter" or other meta data.
    :type matter: Optional[Mapping[str, object]]
    """

    source: str
    filename: str
    uptodate: UpToDate
    matter: Optional[Mapping[str, object]] = None


# flake8: noqa
class BaseLoader(ABC):
    """Base template loader from which all template loaders are derived."""

    def get_source(
        self,
        env: Environment,
        template_name: str,
    ) -> TemplateSource:
        """Get the template source, filename and reload helper for a template"""
        raise NotImplementedError("template loaders must implement a get_source method")

    async def get_source_async(
        self,
        env: Environment,
        template_name: str,
    ) -> TemplateSource:
        """An async version of `get_source`. The default implementation delegates to
        `get_source()`."""
        return self.get_source(env, template_name)

    # pylint: disable=unused-argument
    def get_source_with_context(
        self,
        context: Context,
        template_name: str,
        **kwargs: str,
    ) -> TemplateSource:
        """Get a template's source, optionally referencing a render context."""
        return self.get_source(context.env, template_name)

    # pylint: disable=unused-argument
    async def get_source_with_context_async(
        self,
        context: Context,
        template_name: str,
        **kwargs: str,
    ) -> TemplateSource:
        """An async version of `get_source_with_context`."""
        return await self.get_source_async(context.env, template_name)

    # pylint: disable=redefined-builtin
    def load(
        self,
        env: Environment,
        name: str,
        globals: Optional[Mapping[str, object]] = None,
    ) -> BoundTemplate:
        """Load and parse a template. Used internally by `Environment` to load a
        template source. Delegates to `get_source`.

        A custom loaders would typically implement `get_source` rather than overriding
        `load`.
        """
        try:
            source, filename, uptodate, matter = self.get_source(env, name)
        except Exception as err:
            raise TemplateNotFound(name) from err

        template = env.from_string(
            source,
            globals=globals,
            name=name,
            path=Path(filename),
            matter=matter,
        )
        template.uptodate = uptodate
        return template

    async def load_async(
        self,
        env: Environment,
        name: str,
        globals: Optional[Mapping[str, object]] = None,
    ) -> BoundTemplate:
        """An async version of `load`."""
        try:
            template_source = await self.get_source_async(env, name)
            source, filename, uptodate, matter = template_source
        except Exception as err:
            raise TemplateNotFound(name) from err

        template = env.from_string(
            source,
            globals=globals,
            name=name,
            path=Path(filename),
            matter=matter,
        )
        template.uptodate = uptodate
        return template

    def load_with_context(
        self,
        context: Context,
        name: str,
        **kwargs: str,
    ) -> BoundTemplate:
        """Load and parse a template, optionally referencing a render context."""
        try:
            source, filename, uptodate, matter = self.get_source_with_context(
                context, name, **kwargs
            )
        except Exception as err:
            raise TemplateNotFound(name) from err

        template = context.env.from_string(
            source,
            globals=context.globals,
            name=name,
            path=Path(filename),
            matter=matter,
        )
        template.uptodate = uptodate
        return template

    async def load_with_context_async(
        self,
        context: Context,
        name: str,
        **kwargs: str,
    ) -> BoundTemplate:
        """An async version of `load_with_context`."""
        try:
            (
                source,
                filename,
                uptodate,
                matter,
            ) = await self.get_source_with_context_async(context, name, **kwargs)
        except Exception as err:
            raise TemplateNotFound(name) from err

        template = context.env.from_string(
            source,
            globals=context.globals,
            name=name,
            path=Path(filename),
            matter=matter,
        )
        template.uptodate = uptodate
        return template


class FileSystemLoader(BaseLoader):
    """A loader that loads templates from one or more directories on the file system.

    :param search_path: One or more paths to search.
    :type search_path: Union[str, Path, Iterable[Union[str, Path]]]
    :param encoding: Open template files with the given encoding. Defaults to
        ``"utf-8"``.
    :type encoding: str
    """

    def __init__(
        self,
        search_path: Union[str, Path, Iterable[Union[str, Path]]],
        encoding: str = "utf-8",
    ):
        if not isinstance(search_path, abc.Iterable) or isinstance(search_path, str):
            search_path = [search_path]

        self.search_path = [Path(path) for path in search_path]
        self.encoding = encoding

    def resolve_path(self, template_name: str) -> Path:
        """Return a path to the template `template_name`.

        If the search path is a list of paths, returns the first path where
        `template_name` exists. If none of the search paths contain `template_name`, a
        `TemplateNotFound` exception is raised.
        """
        template_path = Path(template_name)

        if os.path.pardir in template_path.parts:
            raise TemplateNotFound(template_name)

        for path in self.search_path:
            source_path = path.joinpath(template_path)

            if not source_path.exists():
                continue
            return source_path
        raise TemplateNotFound(template_name)

    def _read(self, source_path: Path) -> Tuple[str, float]:
        with source_path.open(encoding=self.encoding) as fd:
            source = fd.read()
        return source, source_path.stat().st_mtime

    def get_source(self, _: Environment, template_name: str) -> TemplateSource:
        source_path = self.resolve_path(template_name)
        source, mtime = self._read(source_path)

        return TemplateSource(
            source, str(source_path), lambda: mtime == source_path.stat().st_mtime
        )

    @staticmethod
    async def _uptodate(source_path: Path, mtime: float) -> bool:
        uptodate = await asyncio.get_running_loop().run_in_executor(
            None, lambda: mtime == source_path.stat().st_mtime
        )
        return uptodate

    async def get_source_async(
        self, env: Environment, template_name: str
    ) -> TemplateSource:
        loop = asyncio.get_running_loop()
        source_path = await loop.run_in_executor(None, self.resolve_path, template_name)
        source, mtime = await loop.run_in_executor(None, self._read, source_path)
        return TemplateSource(
            source, str(source_path), partial(self._uptodate, source_path, mtime)
        )


class FileExtensionLoader(FileSystemLoader):
    """A file system loader that adds a file name extension if one is missing.

    :param search_path: One or more paths to search.
    :type search_path: Union[str, Path, Iterable[Union[str, Path]]]
    :param encoding: Open template files with the given encoding. Defaults to
        ``"utf-8"``.
    :type encoding: str
    :param ext: A default file extension. Should include a leading period. Defaults to
        ``.liquid``.
    :type ext: str
    """

    def __init__(
        self,
        search_path: Union[str, Path, Iterable[Union[str, Path]]],
        encoding: str = "utf-8",
        ext: str = ".liquid",
    ):
        super().__init__(search_path, encoding=encoding)
        self.ext = ext

    def resolve_path(self, template_name: str) -> Path:
        template_path = Path(template_name)

        if not template_path.suffix:
            template_path = template_path.with_suffix(self.ext)

        # Don't allow "../" to escape the search path.
        if os.path.pardir in template_path.parts:
            raise TemplateNotFound(template_name)

        for path in self.search_path:
            source_path = path.joinpath(template_path)

            if not source_path.exists():
                continue
            return source_path
        raise TemplateNotFound(template_name)


class DictLoader(BaseLoader):
    """A loader that loads templates from a dictionary mapping template names to
    template source strings. If the given dictionary is empty,
    :meth:`liquid.Environment.get_template` will always raise a
    :class:`liquid.exceptions.TemplateNotFound` exception.

    :param templates: A dictionary mapping template names to template source strings.
    :type templates: Dict[str, str]
    """

    def __init__(self, templates: Dict[str, str]):
        super().__init__()
        self.templates = templates

    def get_source(self, _: Environment, template_name: str) -> TemplateSource:
        try:
            source = self.templates[template_name]
        except KeyError as err:
            raise TemplateNotFound(template_name) from err

        return TemplateSource(source, template_name, None)


class ChoiceLoader(BaseLoader):
    """A template loader that will try each of a list of loaders until a template is
    found, or raise a :class:`liquid.exceptions.TemplateNotFound` exception if none of
    the loaders could find the template.

    :param loaders: A list of loaders implementing :class:`liquid.loaders.BaseLoader`.
    :type loaders: List[BaseLoader]
    """

    def __init__(self, loaders: List[BaseLoader]):
        super().__init__()
        self.loaders = loaders

    def get_source(self, env: Environment, template_name: str) -> TemplateSource:
        for loader in self.loaders:
            try:
                return loader.get_source(env, template_name)
            except TemplateNotFound:
                pass

        raise TemplateNotFound(template_name)

    async def get_source_async(
        self,
        env: Environment,
        template_name: str,
    ) -> TemplateSource:
        for loader in self.loaders:
            try:
                return await loader.get_source_async(env, template_name)
            except TemplateNotFound:
                pass

        raise TemplateNotFound(template_name)



"""Test Template and Context subclassing."""
import io
import unittest

from typing import Any
from typing import Optional
from typing import List

from liquid import Context
from liquid import Environment

from liquid.context import ContextPath
from liquid.context import Namespace

from liquid.template import BoundTemplate


class CustomTemplateClassTestCase(unittest.TestCase):
    """Test that we can customize the template class."""

    def test_bound_template_subclass(self):
        """Test that we can use a subclass of BoundTemplate with an Environment."""

        class CustomTemplate(BoundTemplate):
            """Mock template subclass"""

        env = Environment()
        env.template_class = CustomTemplate

        template = env.from_string("Hello, {{ you }}!")
        self.assertIsInstance(template, CustomTemplate)

    def test_context_subclass(self):
        """Test that we can use a subclass of Context with an Environment."""

        # pylint: disable=redefined-builtin
        class CustomContext(Context):
            """Mock context subclass"""

            def __init__(
                self,
                env: Environment,
                globals: Optional[Namespace] = None,
                disabled_tags: Optional[List[str]] = None,
                copy_depth: int = 0,
            ):
                super().__init__(env, globals, disabled_tags, copy_depth)
                self.assign_counter = 0

            def assign(self, key: str, val: Any) -> None:
                self.assign_counter += 1
                return super().assign(key, val)

        class CustomTemplate(BoundTemplate):
            """Mock template subclass"""

            context_class = CustomContext

        env = Environment()
        env.template_class = CustomTemplate

        template = env.from_string("{% assign you = 'Brian' %}Hello, {{ you }}!")
        self.assertIsInstance(template, CustomTemplate)

        buf = io.StringIO()
        ctx = CustomContext(env)
        template.render_with_context(ctx, buf)

        self.assertEqual(buf.getvalue(), "Hello, Brian!")
        self.assertEqual(ctx.assign_counter, 1)

    def test_capture_variables_from_context_subclass(self):
        """Test that we can capture a template's used variables from a Context subclass."""

        _missing = object()

        # pylint: disable=redefined-builtin
        class CustomContext(Context):
            """Mock context subclass"""

            def __init__(
                self,
                env: Environment,
                globals: Optional[Namespace] = None,
                disabled_tags: Optional[List[str]] = None,
                copy_depth: int = 0,
            ):
                super().__init__(env, globals, disabled_tags, copy_depth)
                self.references: List[str] = []

            def get(self, path: ContextPath, default: object = ...) -> object:
                self._count_reference(path)
                return super().get(path, default)

            async def get_async(
                self, path: ContextPath, default: object = ...
            ) -> object:
                self._count_reference(path)
                return await super().get_async(path, default)

            def resolve(self, name: str, default: object = _missing) -> Any:
                self._count_reference(name)
                return super().resolve(name, default)

            def _count_reference(self, path: ContextPath) -> None:
                if isinstance(path, str):
                    self.references.append(path)
                else:
                    self.references.append(".".join(str(p) for p in path))

        class CustomTemplate(BoundTemplate):
            """Mock template subclass"""

            context_class = CustomContext

        env = Environment()
        env.template_class = CustomTemplate

        template = env.from_string("{% assign you = 'Brian' %}Hello, {{ you }}!")
        self.assertIsInstance(template, CustomTemplate)

        buf = io.StringIO()
        ctx = CustomContext(env)
        template.render_with_context(ctx, buf)

        self.assertEqual(buf.getvalue(), "Hello, Brian!")
        self.assertEqual(ctx.references, ["you"])

        ctx.resolve("you")
        self.assertEqual(ctx.references, ["you", "you"])



# flake8: noqa
# pylint: disable=useless-import-alias,missing-module-docstring
from liquid.expressions.common import Token
from liquid.expressions.stream import TokenStream

from liquid.expressions.arguments import parse_call_arguments
from liquid.expressions.arguments import parse_keyword_arguments
from liquid.expressions.arguments import parse_macro_arguments
from liquid.expressions.boolean.parse import parse as parse_boolean_expression
from liquid.expressions.boolean.parse import (
    parse_with_parens as parse_boolean_expression_with_parens,
)
from liquid.expressions.conditional.parse import parse as parse_conditional_expression
from liquid.expressions.conditional.parse import (
    parse_with_parens as parse_conditional_expression_with_parens,
)
from liquid.expressions.filtered.parse import parse as parse_filtered_expression
from liquid.expressions.loop.parse import parse as parse_loop_expression


__all__ = (
    "parse_call_arguments",
    "parse_keyword_arguments",
    "parse_macro_arguments",
    "parse_boolean_expression",
    "parse_boolean_expression_with_parens",
    "parse_conditional_expression",
    "parse_conditional_expression_with_parens",
    "parse_filtered_expression",
    "parse_loop_expression",
    "TokenStream",
    "Token",
)
